/**
 * 
 */
package com.wieseke.cptk.input.format.nexus;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.io.IOUtils;

import com.wieseke.cptk.common.api.IDeserializer;
import com.wieseke.cptk.common.api.IPositionHostNode;
import com.wieseke.cptk.common.api.IPositionNode;
import com.wieseke.cptk.common.api.IPositionParasiteNode;
import com.wieseke.cptk.common.dao.PositionNode;
import com.wieseke.cptk.common.dao.ReconstructionCosts;
import com.wieseke.cptk.common.exceptions.DeserializeException;
import com.wieseke.cptk.common.util.NullLogger;
import com.wieseke.cptk.common.util.Tupel;
import com.wieseke.cptk.input.constants.InputFormatNexusConstants;
import com.wieseke.cptk.input.constants.InputViewerConstants;
import com.wieseke.cptk.input.dao.InputCophylogeny;
import com.wieseke.cptk.input.dao.InputHostNode;
import com.wieseke.cptk.input.dao.InputParasiteNode;
import com.wieseke.cptk.reconstruction.dao.EventType;

/**
 * @author Nicolas Wieseke
 * 
 */
public class NexusDeserializer implements IDeserializer {

	/**
	 * Logger
	 */
	// TODO NullLogger
	// static Logger logger = Logger.getLogger(NexusDeserializer.class);
	static NullLogger logger = new NullLogger();
	private InputCophylogeny cophylogeny;

	public InputCophylogeny deserialize(InputStream inputStream, String filename)
			throws DeserializeException {
		logger.debug("---start deserialize from file---");

		String nexusInput = "";
		try {
			StringBuilder sb = new StringBuilder();
			List<?> lines = IOUtils.readLines(inputStream);
			for (Object line : lines) {
				sb.append(line);
				sb.append("\n");
			}
			nexusInput = sb.toString();
			inputStream.close();
		} catch (IOException e) {
			throw new DeserializeException(
					"Error while deserializing nexus file.", e);
		}

		logger.debug("---end deserialize from file---");
		return deserialize(nexusInput, filename);
	}

	public InputCophylogeny deserialize(String nexusInput, String filename)
			throws DeserializeException {
		logger.debug("---start deserialize from string---");
		this.cophylogeny = new InputCophylogeny(filename);

		checkNexusDirective(nexusInput);
		String cleanNexusInput = removeComments(nexusInput);
		Map<String, String> nexusBlocks = getNexusBlocks(cleanNexusInput);

		if (nexusBlocks.containsKey(InputFormatNexusConstants.TAXA_BLOCK_NAME)) {
			handleTaxaBlock(nexusBlocks
					.get(InputFormatNexusConstants.TAXA_BLOCK_NAME));
		} else
			throw new DeserializeException(
					"Error while deserializing nexus file. No TAXA block found in the nexus input data.");
		if (nexusBlocks.containsKey(InputFormatNexusConstants.TREES_BLOCK_NAME)) {
			handleTreesBlock(nexusBlocks
					.get(InputFormatNexusConstants.TREES_BLOCK_NAME));
		} else
			throw new DeserializeException(
					"Error while deserializing nexus file. No TREES block found in the nexus input data.");
		if (nexusBlocks
				.containsKey(InputFormatNexusConstants.COPHYLOGENY_BLOCK_NAME)) {
			handleCophylogenyBlock(nexusBlocks
					.get(InputFormatNexusConstants.COPHYLOGENY_BLOCK_NAME));
		} // else
			// throw new DeserializeException(
			// "Error while deserializing nexus file. No COPHYLOGENY block found in the nexus input data.");

		this.cophylogeny.applyTranslations(this.cophylogeny.getHostRoot());
		this.cophylogeny.applyTranslations(this.cophylogeny.getParasiteRoot());
		this.cophylogeny.applyRanks(this.cophylogeny.getHostRoot());
		this.cophylogeny.applyRanks(this.cophylogeny.getParasiteRoot());
		this.cophylogeny.applyPhi(this.cophylogeny.getParasiteRoot());
		this.cophylogeny.applyReconstruction(this.cophylogeny.getParasiteRoot());
		int unpositionedNodes = this.cophylogeny
				.applyPositions(this.cophylogeny.getHostRoot());
		unpositionedNodes += this.cophylogeny.applyPositions(this.cophylogeny
				.getParasiteRoot());
		int nodeCount = this.cophylogeny.getHostRoot().getChildrenCount()
				+ this.cophylogeny.getParasiteRoot().getChildrenCount() + 2;
		if (((double) nodeCount)
				* InputViewerConstants.UNPOSITIONED_NODE_FRACTION > unpositionedNodes)
			this.cophylogeny.setPositionsApplied(true);
		else
			logger
					.warn("Did not find enough position information for the nodes. Positions will be recalculated.");

		logger.debug("---end deserialize from string---");
		return this.cophylogeny;
	}

	/**
	 * check if the nexus directive is at the beginning of the input data
	 * 
	 * @param nexusInput
	 * @throws DeserializeException
	 */
	private void checkNexusDirective(String nexusInput)
			throws DeserializeException {
		logger.debug("---start checkNexusDirective---");
		String trimmedNexusInput = nexusInput.trim();
		if (trimmedNexusInput.length() < 6
				|| !nexusInput.trim().substring(0, 6).equalsIgnoreCase(
						InputFormatNexusConstants.NEXUS_DIRECTIVE)) {
			throw new DeserializeException(
					"Error while deserializing nexus file. No NEXUS directive found at the beginning of the file.");
		}
		logger.debug("---end checkNexusDirective---");
	}

	/**
	 * remove the comments in the nexus input data
	 * 
	 * @param nexusInput
	 * @return
	 */
	private String removeComments(String nexusInput) {
		logger.debug("---start removeComments---");
		Pattern commentPattern = Pattern
				.compile(InputFormatNexusConstants.COMMENT_PATTERN);
		Matcher commentMatcher = commentPattern.matcher(nexusInput);
		logger.debug("---end removeComments---");
		return commentMatcher.replaceAll("");
	}

	/**
	 * creates a map with the block name as key and block content as value.
	 * 
	 * @param nexusInput
	 * @return
	 */
	private Map<String, String> getNexusBlocks(String nexusInput) {
		logger.debug("---start getNexusBlocks---");
		Map<String, String> nexusBlocks = new HashMap<String, String>();
		Pattern blockPattern = Pattern.compile(
				InputFormatNexusConstants.BLOCK_PATTERN,
				InputFormatNexusConstants.BLOCK_PATTERN_FLAGS);
		Matcher blockMatcher = blockPattern.matcher(nexusInput);
		while (blockMatcher.find()) {
			String key = blockMatcher.group(2).toUpperCase();
			String value = blockMatcher.group(4);

			logger
					.debug("found nexus block " + key + " with content: "
							+ value);
			nexusBlocks.put(key, value);
		}
		logger.debug("---end getNexusBlocks---");
		return nexusBlocks;
	}

	/**
	 * creates a map with the command name as key and content as value.
	 * 
	 * @param nexusInput
	 * @return
	 */
	private List<Tupel<String, String>> getNexusCommands(String blockContent) {
		logger.debug("---start getNexusCommands---");
		List<Tupel<String, String>> nexusCommands = new ArrayList<Tupel<String, String>>();
		Pattern commandPattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_PATTERN,
				InputFormatNexusConstants.COMMAND_PATTERN_FLAGS);
		Matcher commandMatcher = commandPattern.matcher(blockContent);
		while (commandMatcher.find()) {
			String value1 = commandMatcher.group(2);
			String value2 = commandMatcher.group(4);
			logger.debug("found nexus command " + value1 + " with content: "
					+ value2);
			nexusCommands.add(new Tupel<String, String>(value1, value2));
		}
		logger.debug("---end getNexusCommands---");
		return nexusCommands;
	}

	/**
	 * handles the data from the taxa block
	 * 
	 * @param content
	 * @param cophylogeny
	 */
	private void handleTaxaBlock(String content) {
		logger.debug("---start handleTaxaBlock---");
		// not used by the cophylogeny toolkit
		logger.debug("---end handleTaxaBlock---");
	}

	/**
	 * handles the data from the trees block
	 * 
	 * @param content
	 * @param cophylogeny
	 * @throws DeserializeException
	 */
	private void handleTreesBlock(String content) throws DeserializeException {
		logger.debug("---start handleTreesBlock---");
		List<Tupel<String, String>> nexusCommands = getNexusCommands(content);
		for (Tupel<String, String> command : nexusCommands) {
			String commandName = command.getValue1();
			String commandContent = command.getValue2();

			if (commandName
					.equalsIgnoreCase(InputFormatNexusConstants.TREES_BLOCK_TRANSLATE_COMMAND)) {
				handleTreesBlockTranslateCommand(commandContent);
			} else if (commandName
					.equalsIgnoreCase(InputFormatNexusConstants.TREES_BLOCK_TREE_COMMAND)) {
				handleTreesBlockTreeCommand(commandContent);
			}
		}
		if (this.cophylogeny.getHostRoot() == null)
			throw new DeserializeException("No host tree was found");
		if (this.cophylogeny.getParasiteRoot() == null)
			throw new DeserializeException("No parasite tree was found");
		logger.debug("---end handleTreesBlock---");
	}

	/**
	 * create the translations map for the cophylogeny
	 * 
	 * @param commandContent
	 * @throws DeserializeException
	 */
	private void handleTreesBlockTranslateCommand(String commandContent)
			throws DeserializeException {
		logger.debug("---start handleTreesBlockTranslateCommand---");
		Pattern commandListPattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_LIST_PATTERN,
				InputFormatNexusConstants.COMMAND_LIST_PATTERN_FLAGS);
		Matcher commandListMatcher = commandListPattern.matcher(commandContent);
		logger.debug("handle nexus command "
				+ InputFormatNexusConstants.TREES_BLOCK_TRANSLATE_COMMAND);
		Map<String, String> translations = new HashMap<String, String>();
		while (commandListMatcher.find()) {
			logger.debug("found nexus command list item "
					+ commandListMatcher.group(2));
			try {
				String values[] = getCommandListValues(commandListMatcher
						.group(2));
				if (values.length > 2)
					throw new DeserializeException("Command "
							+ commandListMatcher.group(2)
							+ " must consist of exactly 2 values");
				String key = values[0];
				String value = values[1];
				translations.put(key, value);
			} catch (Exception e) {
				throw new DeserializeException("Could not parse the command "
						+ commandListMatcher.group(), e);
			}
		}
		this.cophylogeny.setTranslations(translations);
		logger.debug("---end handleTreesBlockTranslateCommand---");
	}

	/**
	 * create the trees for the cophylogeny
	 * 
	 * @param commandContent
	 */
	private void handleTreesBlockTreeCommand(String commandContent) {
		logger.debug("---start handleTreesBlockTreeCommand---");
		Pattern commandTreePattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_TREE_PATTERN,
				InputFormatNexusConstants.COMMAND_TREE_PATTERN_FLAGS);
		Matcher commandTreeMatcher = commandTreePattern.matcher(commandContent);
		logger.debug("handle nexus command "
				+ InputFormatNexusConstants.TREES_BLOCK_TREE_COMMAND);
		if (commandTreeMatcher.find()) {
			logger.debug("found nexus tree command "
					+ commandTreeMatcher.group());
			String treeName = commandTreeMatcher.group(4);
			String treeString = commandTreeMatcher.group(8).replaceAll(
					"[\b\t\f\n\r]", "");
			if (treeName
					.equalsIgnoreCase(InputFormatNexusConstants.TREES_BLOCK_HOST_TREE_NAME)) {
				IPositionNode hostRoot = parseTree(treeString, 0, true);
				this.cophylogeny.createHostTree((IPositionHostNode) hostRoot);
			} else if (treeName
					.equalsIgnoreCase(InputFormatNexusConstants.TREES_BLOCK_PARASITE_TREE_NAME)) {
				IPositionNode parasiteRoot = parseTree(treeString, 0, false);
				this.cophylogeny
						.createParasiteTree((IPositionParasiteNode) parasiteRoot);
			} else
				logger
						.warn("found tree with name different to 'HOST' or 'PARASITE'. This tree will not be used.");
		}
		logger.debug("---end handleTreesBlockTreeCommand---");
	}

	/**
	 * parses the tree string
	 * 
	 * @param treeString
	 * @return the corresponding InputTree
	 */
	private PositionNode parseTree(String treeString, int nodeNumber,
			boolean isHostTree) {
		logger.debug("---start parseTree---");
		PositionNode node;
		if (isHostTree)
			node = new InputHostNode();
		else
			node = new InputParasiteNode();
		int startIndex = 0;
		int endIndex = -1;
		int bracketCount = 0;
		int nextNodeNumber = nodeNumber + 1;

		for (int index = 0; index < treeString.length(); index++) {
			if (treeString.charAt(index) == '(') {
				if (bracketCount == 0) {
					startIndex = index + 1;
				}
				bracketCount++;
			} else if (treeString.charAt(index) == ')') {
				bracketCount--;
				if (bracketCount == 0) {
					endIndex = index;
					PositionNode child = parseTree(treeString.substring(
							startIndex, endIndex), nextNodeNumber, isHostTree);
					nextNodeNumber += child.getChildrenCount() + 1;
					child.setParent(node);
					node.addChild(child);
					startIndex = index + 1;
					break;
				}
			} else if (treeString.charAt(index) == ',') {
				if (bracketCount == 1) {
					endIndex = index;
					PositionNode child = parseTree(treeString.substring(
							startIndex, endIndex), nextNodeNumber, isHostTree);
					nextNodeNumber += child.getChildrenCount() + 1;
					child.setParent(node);
					node.addChild(child);
					startIndex = index + 1;
				}
			}
		}
		node.setNumber(nodeNumber);
		String label = treeString.substring(endIndex + 1);
		if (label.equals("")) {
			label = (isHostTree ? "H" + nodeNumber : "P" + nodeNumber);
		}
		node.setLabel(label);
		int childrenCount = node.getChildren().size();
		for (IPositionNode child : node.getChildren()) {
			childrenCount += child.getChildrenCount();
		}
		node.setChildrenCount(childrenCount);
		logger.debug("---end parseTree---");
		return node;
	}

	/**
	 * handles the data from the cophylogeny block
	 * 
	 * @param content
	 * @param cophylogeny
	 * @throws DeserializeException
	 */
	private void handleCophylogenyBlock(String content)
			throws DeserializeException {
		logger.debug("---start handleCophylogenyBlock---");
		List<Tupel<String, String>> nexusCommands = getNexusCommands(content);
		for (Tupel<String, String> command : nexusCommands) {
			String commandName = command.getValue1();
			String commandContent = command.getValue2();

			if (commandName
					.equalsIgnoreCase(InputFormatNexusConstants.COPHYLOGENY_BLOCK_RANKS_COMMAND)) {
				handleCophylogenyBlockRanksCommand(commandContent);
			} else if (commandName
					.equalsIgnoreCase(InputFormatNexusConstants.COPHYLOGENY_BLOCK_PHI_COMMAND)) {
				handleCophylogenyBlockPhiCommand(commandContent);
			} else if (commandName
					.equalsIgnoreCase(InputFormatNexusConstants.COPHYLOGENY_BLOCK_RECONSTRUCTIONEVENTS_COMMAND)) {
				handleCophylogenyBlockReconstructionEventsCommand(commandContent);
			} else if (commandName
					.equalsIgnoreCase(InputFormatNexusConstants.COPHYLOGENY_BLOCK_RECONSTRUCTION_COMMAND)) {
				handleCophylogenyBlockReconstructionCommand(commandContent);
			} else if (commandName
					.equalsIgnoreCase(InputFormatNexusConstants.COPHYLOGENY_BLOCK_POSITIONS_COMMAND)) {
				handleCophylogenyBlockPositionsCommand(commandContent);
			} else if (commandName
					.equalsIgnoreCase(InputFormatNexusConstants.COPHYLOGENY_BLOCK_COSTS_COMMAND)) {
				handleCophylogenyBlockCostCommand(commandContent);
			} else if (commandName
					.equalsIgnoreCase(InputFormatNexusConstants.COPHYLOGENY_BLOCK_OPTIONS_COMMAND)) {
				handleCophylogenyBlockOptionsCommand(commandContent);
			}
		}
		logger.debug("---end handleCophylogenyBlock---");
	}

	/**
	 * create the ranks map for the cophylogeny
	 * 
	 * @param commandContent
	 * @throws DeserializeException
	 */
	private void handleCophylogenyBlockRanksCommand(String commandContent)
			throws DeserializeException {
		logger.debug("---start handleCophylogenyBlockRanksCommand---");
		Pattern commandListPattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_LIST_PATTERN,
				InputFormatNexusConstants.COMMAND_LIST_PATTERN_FLAGS);
		Matcher commandListMatcher = commandListPattern.matcher(commandContent);
		logger.debug("handle nexus command "
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_RANKS_COMMAND);
		Map<String, Tupel<Integer, Integer>> ranks = new LinkedHashMap<String, Tupel<Integer, Integer>>();
		while (commandListMatcher.find()) {
			logger.debug("found nexus command list item "
					+ commandListMatcher.group(2));
			String key;
			Integer rankFrom;
			Integer rankTo;
			try {
				String values[] = getCommandListValues(commandListMatcher
						.group(2));
				if (values.length != 3)
					throw new DeserializeException("Command "
							+ commandListMatcher.group(2)
							+ " must consist of exactly 3 values");
				key = values[0];
				rankFrom = Integer.parseInt(values[1]);
				try {
					rankTo = Integer.parseInt(values[2]);
				} catch (Exception e) {
					logger
							.warn("Could not parse the second rank value in command "
									+ commandListMatcher.group()
									+ ". Will use the first rank value instead.");
					rankTo = rankFrom = Integer.parseInt(values[1]);
				}
				ranks.put(key, new Tupel<Integer, Integer>(rankFrom, rankTo));
			} catch (NumberFormatException e) {
				throw new DeserializeException("Could not parse the command "
						+ commandListMatcher.group(), e);
			}
		}
		this.cophylogeny.setRanks(ranks);
		logger.debug("---end handleCophylogenyBlockRanksCommand---");
	}

	/**
	 * create the phi association map for the cophylogeny
	 * 
	 * @param commandContent
	 * @throws DeserializeException
	 */
	private void handleCophylogenyBlockPhiCommand(String commandContent)
			throws DeserializeException {
		logger.debug("---start handleCophylogenyBlockPhiCommand---");
		Pattern commandListPattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_LIST_PATTERN,
				InputFormatNexusConstants.COMMAND_LIST_PATTERN_FLAGS);
		Matcher commandListMatcher = commandListPattern.matcher(commandContent);
		logger.debug("handle nexus command "
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_PHI_COMMAND);
		Map<String, ArrayList<String>> phi = new LinkedHashMap<String, ArrayList<String>>();
		while (commandListMatcher.find()) {
			logger.debug("found nexus command list item "
					+ commandListMatcher.group(2));
			try {
				String values[] = getCommandListValues(commandListMatcher
						.group(2));
				if (values.length != 2)
					throw new DeserializeException("Command "
							+ commandListMatcher.group(2)
							+ " must consist of exactly 2 values");
				String parasite = values[0];
				String host = values[1];
				ArrayList<String> hostList = phi.get(parasite);
				if (hostList == null)
					hostList = new ArrayList<String>();
				hostList.add(host);
				phi.put(parasite, hostList);
			} catch (Exception e) {
				throw new DeserializeException("Could not parse the command "
						+ commandListMatcher.group(), e);
			}
		}
		this.cophylogeny.setPhi(phi);
		logger.debug("---end handleCophylogenyBlockPhiCommand---");
	}

	private void handleCophylogenyBlockReconstructionEventsCommand(
			String commandContent) throws DeserializeException {
		logger.debug("---start handleCophylogenyBlockReconstructionEventsCommand---");
		Pattern commandListPattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_LIST_PATTERN,
				InputFormatNexusConstants.COMMAND_LIST_PATTERN_FLAGS);
		Matcher commandListMatcher = commandListPattern.matcher(commandContent);
		logger.debug("handle nexus command "
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_RECONSTRUCTIONEVENTS_COMMAND);
		EventType events = new EventType(-1, -1, -1, -1);
		while (commandListMatcher.find()) {
			logger.debug("found nexus command list item "
					+ commandListMatcher.group(2));
			try {
				String values[] = getCommandListValues(commandListMatcher
						.group(2));
				if (values.length != 2)
					throw new DeserializeException("Command "
							+ commandListMatcher.group(2)
							+ " must consist of exactly 2 values");
				String eventname = values[0].toUpperCase();
				Integer value = new Integer(values[1]);
				events.setEvent(eventname, value);
			} catch (NumberFormatException e) {
				throw new DeserializeException("Could not parse the command "
						+ commandListMatcher.group(), e);
			}
		}
		this.cophylogeny.setReconstructionEvents(events);
		logger.debug("---end handleCophylogenyBlockReconstructionEventsCommand---");
	}

	/**
	 * create the reconstruction association map for the cophylogeny
	 * 
	 * @param commandContent
	 * @throws DeserializeException
	 */
	private void handleCophylogenyBlockReconstructionCommand(
			String commandContent) throws DeserializeException {
		logger.debug("---start handleCophylogenyBlockReconstructionCommand---");
		Pattern commandListPattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_LIST_PATTERN,
				InputFormatNexusConstants.COMMAND_LIST_PATTERN_FLAGS);
		Matcher commandListMatcher = commandListPattern.matcher(commandContent);
		logger.debug("handle nexus command "
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_RECONSTRUCTION_COMMAND);
		Map<String, ArrayList<String>> reconstruction = new LinkedHashMap<String, ArrayList<String>>();
		while (commandListMatcher.find()) {
			logger.debug("found nexus command list item "
					+ commandListMatcher.group(2));
			try {
				String values[] = getCommandListValues(commandListMatcher
						.group(2));
				if (values.length != 2)
					throw new DeserializeException("Command "
							+ commandListMatcher.group(2)
							+ " must consist of exactly 2 values");
				String parasite = values[0];
				String host = values[1];
				ArrayList<String> hostList = reconstruction.get(parasite);
				if (hostList == null)
					hostList = new ArrayList<String>();
				hostList.add(host);
				reconstruction.put(parasite, hostList);
			} catch (Exception e) {
				throw new DeserializeException("Could not parse the command "
						+ commandListMatcher.group(), e);
			}
		}
		this.cophylogeny.setReconstruction(reconstruction);
		logger.debug("---end handleCophylogenyBlockReconstructionCommand---");
	}

	/**
	 * create the phi association map for the cophylogeny
	 * 
	 * @param commandContent
	 * @throws DeserializeException
	 */
	private void handleCophylogenyBlockPositionsCommand(String commandContent)
			throws DeserializeException {
		logger.debug("---start handleCophylogenyBlockPositionsCommand---");
		Pattern commandListPattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_LIST_PATTERN,
				InputFormatNexusConstants.COMMAND_LIST_PATTERN_FLAGS);
		Matcher commandListMatcher = commandListPattern.matcher(commandContent);
		logger.debug("handle nexus command "
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_PHI_COMMAND);
		Map<String, Tupel<Integer, Integer>> positions = new LinkedHashMap<String, Tupel<Integer, Integer>>();
		while (commandListMatcher.find()) {
			logger.debug("found nexus command list item "
					+ commandListMatcher.group(2));
			try {
				String values[] = getCommandListValues(commandListMatcher
						.group(2));
				if (values.length != 3)
					throw new DeserializeException("Command "
							+ commandListMatcher.group(2)
							+ " must consist of exactly 3 values");
				String node = values[0];
				int posX = Integer.valueOf(values[1].trim());
				int posY = Integer.valueOf(values[2].trim());
				positions.put(node, new Tupel<Integer, Integer>(posX, posY));
			} catch (Exception e) {
				throw new DeserializeException("Could not parse the command "
						+ commandListMatcher.group(), e);
			}
		}
		this.cophylogeny.setPositions(positions);
		logger.debug("---end handleCophylogenyBlockPositionsCommand---");
	}

	/**
	 * create the cost map for the cophylogeny
	 * 
	 * @param commandContent
	 * @throws DeserializeException
	 */
	private void handleCophylogenyBlockCostCommand(String commandContent)
			throws DeserializeException {
		logger.debug("---start handleCophylogenyBlockCostCommand---");
		Pattern commandListPattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_LIST_PATTERN,
				InputFormatNexusConstants.COMMAND_LIST_PATTERN_FLAGS);
		Matcher commandListMatcher = commandListPattern.matcher(commandContent);
		logger.debug("handle nexus command "
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_COSTS_COMMAND);
		ReconstructionCosts costs = InputCophylogeny.createStandardCosts();
		while (commandListMatcher.find()) {
			logger.debug("found nexus command list item "
					+ commandListMatcher.group(2));
			try {
				String values[] = getCommandListValues(commandListMatcher
						.group(2));
				if (values.length != 2)
					throw new DeserializeException("Command "
							+ commandListMatcher.group(2)
							+ " must consist of exactly 2 values");
				String costname = values[0].toUpperCase();
				BigDecimal value = new BigDecimal(values[1]);
				costs.put(costname, value);
			} catch (NumberFormatException e) {
				throw new DeserializeException("Could not parse the command "
						+ commandListMatcher.group(), e);
			}
		}
		this.cophylogeny.setCosts(costs);
		logger.debug("---end handleCophylogenyBlockCostCommand---");
	}

	/**
	 * create the options map for the cophylogeny
	 * 
	 * @param commandContent
	 * @throws DeserializeException
	 */
	private void handleCophylogenyBlockOptionsCommand(String commandContent)
			throws DeserializeException {
		logger.debug("---start handleCophylogenyBlockOptionsCommand---");
		Pattern commandListPattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_LIST_PATTERN,
				InputFormatNexusConstants.COMMAND_LIST_PATTERN_FLAGS);
		Matcher commandListMatcher = commandListPattern.matcher(commandContent);
		logger.debug("handle nexus command "
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_OPTIONS_COMMAND);
		Map<String, String> options = InputCophylogeny.createStandardOptions();
		while (commandListMatcher.find()) {
			logger.debug("found nexus command list item "
					+ commandListMatcher.group(2));
			try {
				String values[] = getCommandListValues(commandListMatcher
						.group(2));
				if (values.length != 2)
					throw new DeserializeException("Command "
							+ commandListMatcher.group(2)
							+ " must consist of exactly 2 values");
				String optionname = values[0].toUpperCase();
				String value = values[1];
				options.put(optionname, value);
			} catch (NumberFormatException e) {
				throw new DeserializeException("Could not parse the command "
						+ commandListMatcher.group(), e);
			}
		}
		this.cophylogeny.setOptions(options);
		logger.debug("---end handleCophylogenyBlockOptionsCommand---");
	}

	private String[] getCommandListValues(String group) {
		ArrayList<String> values = new ArrayList<String>();

		Pattern commandListValuePattern = Pattern.compile(
				InputFormatNexusConstants.COMMAND_LIST_VALUE_PATTERN,
				InputFormatNexusConstants.COMMAND_LIST_VALUE_PATTERN_FLAGS);
		Matcher commandListValueMatcher = commandListValuePattern
				.matcher(group);
		while (commandListValueMatcher.find()) {
			String value = commandListValueMatcher.group();
			if (value.startsWith("'")) {
				value = value.substring(1);
			}
			if (value.endsWith("'")) {
				value = value.substring(0, value.length() - 1);
			}
			value = value.replaceAll("''", "'");
			values.add(value);
		}
		String strValues[] = new String[values.size()];
		int i = 0;
		for (String strValue : values) {
			strValues[i] = strValue;
			i++;
		}
		return strValues;
	}
}
