package com.wieseke.cptk.input.layout;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.gnu.glpk.GLPK;
import org.gnu.glpk.GLPKConstants;
import org.gnu.glpk.SWIGTYPE_p_double;
import org.gnu.glpk.SWIGTYPE_p_int;
import org.gnu.glpk.glp_iocp;
import org.gnu.glpk.glp_prob;

import com.wieseke.cptk.common.api.IPositionNode;
import com.wieseke.cptk.common.api.ITreeArrangement;
import com.wieseke.cptk.common.log.CophylogenyLogger;
import com.wieseke.cptk.common.util.NodeUtils;
import com.wieseke.cptk.input.constants.InputViewerConstants;
import com.wieseke.cptk.input.dao.InputHostNode;
import com.wieseke.cptk.input.dao.InputNode;
import com.wieseke.cptk.input.dao.InputParasiteNode;

@SuppressWarnings("unused")
public class LayoutFormatter {

	static {
		try {
			if (System.getProperty("os.name").toLowerCase().contains("windows")) {
				// try to load Windows library
				System.loadLibrary("glpk_4_47");
				System.loadLibrary("glpk_4_47_java");
			} else {
				// try to load Linux library
				System.loadLibrary("glpk");
				System.loadLibrary("glpk_java");
			}
		} catch (UnsatisfiedLinkError e) {
			CophylogenyLogger
					.logError(
							"The dynamic link library for GLPK for Java could not be"
									+ "loaded.\nConsider using\njava -Djava.library.path=",
							e);
		}
	}

	/**
	 * @param hostRoot
	 *            - the root node of the host tree
	 * @param parasiteRoot
	 *            - the root node of the parasite tree
	 * @param arrangement
	 *            - layout arrangement information
	 */
	public static void recalculateNodeOrder(InputHostNode hostRoot,
			InputParasiteNode parasiteRoot, ITreeArrangement arrangement) {
		GLPKSolver(hostRoot, parasiteRoot, arrangement);
		// bruteForce(hostRoot, parasiteRoot, arrangement);
	}

	public static void GLPKSolver(InputHostNode hostRoot,
			InputParasiteNode parasiteRoot, ITreeArrangement arrangement) {
		/**
		 * Initialization of variables used
		 */
		int internalNodesHost, internalNodesParasite, countInternalParaNodes = 0;
		int countInternalHostNodes = 0;
		HashMap<InputHostNode, Integer> hostMap = new HashMap<InputHostNode, Integer>();
		HashMap<InputParasiteNode, Integer> paraMap = new HashMap<InputParasiteNode, Integer>();

		internalNodesHost = getNrOfInternalNodes(hostRoot);
		internalNodesParasite = getNrOfInternalNodes(parasiteRoot);

		ArrayList<InputHostNode> hostleaves;
		ArrayList<InputParasiteNode> parasiteleaves;
		hostleaves = new ArrayList<InputHostNode>();
		parasiteleaves = new ArrayList<InputParasiteNode>();
		findHostleaves(hostRoot, hostleaves);
		findParasiteleaves(parasiteRoot, parasiteleaves);

		int[][] weightsolid, weightdashed;
		// weightsolid and weightdashed are matrices to represent edges between
		// two nodes in the graph. The integer in entry (i,j) represents how
		// many edges are drawn from node i to j.
		// 1. coordinate is for host, and 2. is for parasite
		weightsolid = new int[internalNodesHost][internalNodesParasite];
		weightdashed = new int[internalNodesHost][internalNodesParasite];

		InputParasiteNode paraleaf1, paraleaf2;
		List<InputHostNode> ass1, ass2;
		InputHostNode ancHost;
		InputParasiteNode ancPara;

		InputHostNode host1, host2;

		boolean crossed;

		/**
		 * This forloop creates the digraph, represented by weightsolid and
		 * weightdashed. Further we map each internal nodes name (label) to the
		 * index in the matrices (hostMap and paraMap).
		 */

		int nrcrossescount = 0;

		for (int i = 0; i < parasiteleaves.size(); i++) {
			paraleaf1 = parasiteleaves.get(i);
			ass1 = paraleaf1.getAssociations();
			for (int j = i + 1; j < parasiteleaves.size(); j++) {
				paraleaf2 = parasiteleaves.get(j);
				ass2 = paraleaf2.getAssociations();
				ancPara = (InputParasiteNode) NodeUtils.getCommonAncestor(
						paraleaf1, paraleaf2);
				if (!paraMap.containsKey(ancPara)) {
					paraMap.put(ancPara, countInternalParaNodes);
					countInternalParaNodes++;
				}
				for (int ass1count = 0; ass1count < ass1.size(); ass1count++) {
					host1 = ass1.get(ass1count);
					for (int ass2count = 0; ass2count < ass2.size(); ass2count++) {
						host2 = ass2.get(ass2count);
						if (host1.equals(host2))
							continue;
						ancHost = (InputHostNode) NodeUtils.getCommonAncestor(
								host1, host2);
						if (!hostMap.containsKey(ancHost)) {
							hostMap.put(ancHost, countInternalHostNodes);
							countInternalHostNodes++;
						}
						crossed = isCrossed(paraleaf1, host1, paraleaf2, host2,
								arrangement);

						if (crossed) {
							weightdashed[hostMap.get(ancHost)][paraMap
									.get(ancPara)]++;
							nrcrossescount++;
						} else {
							weightsolid[hostMap.get(ancHost)][paraMap
									.get(ancPara)]++;
						}

					}
				}
			}
		}

		/**
		 * numedges is the number of edges in the digraph
		 */
		int numedges = 0;
		for (InputHostNode h : hostMap.keySet()) {
			for (InputParasiteNode p : paraMap.keySet()) {
				if (weightdashed[hostMap.get(h)][paraMap.get(p)] != 0
						|| weightsolid[hostMap.get(h)][paraMap.get(p)] != 0) {
					numedges++;
				}
			}
		}

		/**
		 * Here we initialise the GLPK instance.
		 */

		// create a glpk problem
		glp_prob lp = GLPK.glp_create_prob();
		GLPK.glp_set_prob_name(lp, "balanced_subgraph");

		// add variables (columns)
		GLPK.glp_add_cols(lp, numedges + internalNodesHost
				+ internalNodesParasite);

		for (InputHostNode h : hostMap.keySet()) {
			GLPK.glp_set_col_name(lp, hostMap.get(h) + 1, h.getLabel());
			GLPK.glp_set_col_kind(lp, hostMap.get(h) + 1, GLPKConstants.GLP_BV);
			GLPK.glp_set_col_bnds(lp, hostMap.get(h) + 1, GLPKConstants.GLP_DB,
					0, 1);
		}
		for (InputParasiteNode p : paraMap.keySet()) {
			GLPK.glp_set_col_name(lp, internalNodesHost + paraMap.get(p) + 1, p
					.getLabel());
			GLPK.glp_set_col_kind(lp, internalNodesHost + paraMap.get(p) + 1,
					GLPKConstants.GLP_BV);
			GLPK.glp_set_col_bnds(lp, internalNodesHost + paraMap.get(p) + 1,
					GLPKConstants.GLP_DB, 0, 1);
		}
		for (int j = 0; j < numedges; j++) {
			GLPK.glp_set_col_name(lp, internalNodesHost + internalNodesParasite
					+ 1 + j, "k" + j);
			GLPK.glp_set_col_kind(lp, internalNodesHost + internalNodesParasite
					+ 1 + j, GLPKConstants.GLP_IV);
			GLPK.glp_set_col_bnds(lp, internalNodesHost + internalNodesParasite
					+ 1 + j, GLPKConstants.GLP_DB, 0, 2);
		}

		int[] sums = new int[numedges + internalNodesHost
				+ internalNodesParasite + 1];

		// add constraints (rows)
		GLPK.glp_add_rows(lp, numedges);
		// ia: array of row indices
		SWIGTYPE_p_int ia;
		// ja: array of column indices
		SWIGTYPE_p_int ja;
		// ar: value of the coefficient
		SWIGTYPE_p_double ar;
		ia = GLPK.new_intArray(1 + 3 * numedges);
		ja = GLPK.new_intArray(1 + 3 * numedges);
		ar = GLPK.new_doubleArray(1 + 3 * numedges);

		for (int i = 1; i <= numedges * 3; i++) {
			GLPK.intArray_setitem(ia, i, (i - 1) / 3 + 1);
		}

		/**
		 * Removes edges where there are both dashed and solid edges between the
		 * same pair of vertices. (We remove the least weighted) Here we also
		 * set some necessary variables in GLPK to the specific problem.
		 */
		boolean solid;
		int weight;
		int min, numDashed, numSolid;
		int rowcount = 0;
		for (InputHostNode h : hostMap.keySet()) {
			for (InputParasiteNode p : paraMap.keySet()) {
				numDashed = weightdashed[hostMap.get(h)][paraMap.get(p)];
				numSolid = weightsolid[hostMap.get(h)][paraMap.get(p)];
				if (numDashed > 0 || numSolid > 0) {
					GLPK.intArray_setitem(ja, rowcount * 3 + 1,
							hostMap.get(h) + 1);
					GLPK.intArray_setitem(ja, rowcount * 3 + 2,
							internalNodesHost + paraMap.get(p) + 1);
					GLPK.intArray_setitem(ja, rowcount * 3 + 3,
							internalNodesHost + internalNodesParasite + 1
									+ rowcount);
					if (numSolid > 0 && numDashed > 0) {
						min = Math.min(numDashed, numSolid);
						weightdashed[hostMap.get(h)][paraMap.get(p)] -= min;
						weightsolid[hostMap.get(h)][paraMap.get(p)] -= min;
					}
					solid = weightdashed[hostMap.get(h)][paraMap.get(p)] < weightsolid[hostMap
							.get(h)][paraMap.get(p)];
					weight = Math.max(weightdashed[hostMap.get(h)][paraMap.get(p)], weightsolid[hostMap.get(h)][paraMap.get(p)]);
					GLPK.glp_set_row_name(lp, rowcount + 1, "c" + rowcount);
					GLPK.glp_set_row_bnds(lp, rowcount + 1,
							GLPKConstants.GLP_LO, (solid ? 0 : 1), 42);

					GLPK.doubleArray_setitem(ar, rowcount * 3 + 1, -1);
					GLPK.doubleArray_setitem(ar, rowcount * 3 + 2, -1);
					GLPK.doubleArray_setitem(ar, rowcount * 3 + 3, 2);

					sums[hostMap.get(h) + 1] -= weight;
					sums[internalNodesHost + paraMap.get(p) + 1] -= weight;
					sums[internalNodesHost + internalNodesParasite + 1
							+ rowcount] = 2 * weight;
					rowcount++;
				}
			}
		}

		// load the date into the mip coefficients matrix
		GLPK.glp_load_matrix(lp, 3 * numedges, ia, ja, ar);

		// set the objective function
		// this will be the sum of all rows

		GLPK.glp_set_obj_name(lp, "sum");
		GLPK.glp_set_obj_dir(lp, GLPKConstants.GLP_MIN);

		for (int i = 1; i < sums.length; i++) {
			GLPK.glp_set_obj_coef(lp, i, sums[i]);
		}

		// initialize mip properties
		glp_iocp iocp = new glp_iocp();
		GLPK.glp_init_iocp(iocp);
		// set presolve property
		iocp.setPresolve(GLPKConstants.GLP_ON);
		// set timeout
		iocp.setTm_lim(5000);

		// solve
		int ret = GLPK.glp_intopt(lp, iocp);

		if (ret == 0) {
			int status = GLPK.glp_mip_status(lp);
			if (status == GLPKConstants.GLP_UNDEF) {
				// MIP solution is undefined.
			} else if (status == GLPKConstants.GLP_OPT) {
				// MIP solution is integer optimal.
			} else if (status == GLPKConstants.GLP_FEAS) {
				// MIP solution is integer feasible, however, its optimality (or
				// non-optimality) has not been proven, perhaps due to premature
				// termination of the search.
			} else if (status == GLPKConstants.GLP_NOFEAS) {
				// problem has no integer feasible solution (proven by the
				// solver).
			}
		}
		int val;
		for (InputHostNode h : hostMap.keySet()) {
			val = (int) GLPK.glp_mip_col_val(lp, hostMap.get(h) + 1);
			if (val == 1) {
				swapChildOrder(h);
			}
		}
		for (InputParasiteNode p : paraMap.keySet()) {
			val = (int) GLPK.glp_mip_col_val(lp, internalNodesHost
					+ paraMap.get(p) + 1);
			if (val == 1) {
				swapChildOrder(p);
			}
		}
		// free memory
		GLPK.glp_delete_prob(lp);

	}

	@SuppressWarnings("unchecked")
	public static int getNrOfInternalNodes(InputNode i) {
		if (i.isLeaf()) {
			return 0;
		}
		List<InputNode> children = (List<InputNode>) i.getChildren();
		int c = 0;
		for (InputNode child : children) {
			c += getNrOfInternalNodes(child);
		}
		return 1 + c;
	}

	public static void findHostleaves(InputHostNode node,
			List<InputHostNode> leaves) {
		List<InputHostNode> children;
		if (node.isLeaf()) {
			leaves.add(node);
		} else {
			children = node.getChildren();
			for (int i = 0; i < children.size(); i++) {
				findHostleaves(children.get(i), leaves);
			}
		}
	}

	/**
	 * Adds leaves to the list leaves
	 * */

	public static void findParasiteleaves(InputParasiteNode node,
			List<InputParasiteNode> leaves) {
		List<InputParasiteNode> children;
		if (node.isLeaf()) {
			leaves.add(node);
		} else {
			children = node.getChildren();
			for (int i = 0; i < children.size(); i++) {
				findParasiteleaves(children.get(i), leaves);
			}
		}

	}

	static int combies, number, bestCol;
	// 'res' is the list of the best coloring.
	static int[] res;

	/*
	 * Computes all color combinations and finds the best one, given the least
	 * number of color violations in the digraph.
	 */

	public static void bruteForce(InputHostNode hostRoot,
			InputParasiteNode parasiteRoot, ITreeArrangement arrangement) {
		bestCol = Integer.MAX_VALUE;
		number = 0;
		int internalNodesHost, internalNodesParasite, countInternalParaNodes = 0;
		int countInternalHostNodes = 0;
		HashMap<InputHostNode, Integer> hostMap = new HashMap<InputHostNode, Integer>();
		HashMap<InputParasiteNode, Integer> paraMap = new HashMap<InputParasiteNode, Integer>();

		internalNodesHost = getNrOfInternalNodes(hostRoot);
		internalNodesParasite = getNrOfInternalNodes(parasiteRoot);

		ArrayList<InputHostNode> hostleaves;
		ArrayList<InputParasiteNode> parasiteleaves;
		hostleaves = new ArrayList<InputHostNode>();
		parasiteleaves = new ArrayList<InputParasiteNode>();
		findHostleaves(hostRoot, hostleaves);

		findParasiteleaves(parasiteRoot, parasiteleaves);

		int[][] weightsolid, weightdashed;
		// weightsolid and weightdashed are matrices to represent edges between
		// two nodes in the graph. The integer in entry (i,j) represents how
		// many edges are drawn from node i to j.
		// 1. coordinate is for host, and 2. is for parasite
		weightsolid = new int[internalNodesHost][internalNodesParasite];
		weightdashed = new int[internalNodesHost][internalNodesParasite];

		InputParasiteNode paraleaf1, paraleaf2;
		List<InputHostNode> ass1, ass2;
		InputHostNode ancHost;
		InputParasiteNode ancPara;

		InputHostNode host1, host2;

		boolean crossed;

		/**
		 * This forloop creates the digraph, represented by weightsolid and
		 * weightdashed. Further we map each internal nodes name (label) to the
		 * index in the matrices (hostMap and paraMap).
		 */

		int nrcrossescount = 0;

		for (int i = 0; i < parasiteleaves.size(); i++) {
			paraleaf1 = parasiteleaves.get(i);
			ass1 = paraleaf1.getAssociations();
			for (int j = i + 1; j < parasiteleaves.size(); j++) {
				paraleaf2 = parasiteleaves.get(j);
				ass2 = paraleaf2.getAssociations();
				ancPara = (InputParasiteNode) NodeUtils.getCommonAncestor(
						paraleaf1, paraleaf2);
				if (!paraMap.containsKey(ancPara)) {
					paraMap.put(ancPara, countInternalParaNodes);
					countInternalParaNodes++;
				}
				for (int ass1count = 0; ass1count < ass1.size(); ass1count++) {
					host1 = ass1.get(ass1count);
					for (int ass2count = 0; ass2count < ass2.size(); ass2count++) {
						host2 = ass2.get(ass2count);
						ancHost = (InputHostNode) NodeUtils.getCommonAncestor(
								host1, host2);
						if (!hostMap.containsKey(ancHost)) {
							hostMap.put(ancHost, countInternalHostNodes);
							countInternalHostNodes++;
						}
						crossed = isCrossed(paraleaf1, host1, paraleaf2, host2,
								arrangement);

						if (crossed) {
							weightdashed[hostMap.get(ancHost)][paraMap
									.get(ancPara)]++;
							nrcrossescount++;
						} else {
							weightsolid[hostMap.get(ancHost)][paraMap
									.get(ancPara)]++;
						}
					}
				}
			}
		}

		InputNode[] nodes = new InputNode[hostMap.size() + paraMap.size()];

		int count = 0;
		for (InputHostNode h : hostMap.keySet()) {
			nodes[count] = h;
			count++;
		}
		for (InputParasiteNode p : paraMap.keySet()) {
			nodes[count] = p;
			count++;
		}

		int numhost = hostMap.size();
		boolean stop = false;
		combies = (int) Math.pow(2,
				(weightdashed.length + weightdashed[0].length));
		int crossings = 0;
		fill(new int[weightdashed.length + weightdashed[0].length], stop,
				numhost, nodes, weightsolid, weightdashed, hostMap, paraMap);

		/*
		 * Swap the necessary nodes
		 */
		for (int i = 0; i < nodes.length; i++) {
			if (res[i] == 1) {
				swapChildOrder(nodes[i]);
			}
		}
	}

	/*
	 * The next two methods provides the permutations of colors to the vertices
	 * of the digraph, and finds the one leading to the smallest number of
	 * violations.
	 */
	public static void fill(int[] array, boolean stop, int numhost,
			InputNode[] nodes, int[][] weightsolid, int[][] weightdashed,
			HashMap<InputHostNode, Integer> hostMap,
			HashMap<InputParasiteNode, Integer> paraMap) {
		if (stop)
			return;
		fill(array, 0, stop, numhost, nodes, weightsolid, weightdashed,
				hostMap, paraMap);
	}

	private static void fill(int[] array, int n, boolean stop, int numhost,
			InputNode[] nodes, int[][] weightsolid, int[][] weightdashed,
			HashMap<InputHostNode, Integer> hostMap,
			HashMap<InputParasiteNode, Integer> paraMap) {
		if (n >= array.length) {
			number++;
			/*
			 * Tjek if this combination leads to less violations. If yes, update
			 * the best found solution to this solution.
			 */
			int temp = numViolations(array, numhost, nodes, weightsolid,
					weightdashed, hostMap, paraMap);
			if (temp < bestCol) {
				bestCol = temp;
				res = array.clone();
			}
			if (number == (int) combies) {
				stop = true;
				return;
			}
		} else {
			if (stop)
				return;
			array[n] = 1;
			fill(array, n + 1, stop, numhost, nodes, weightsolid, weightdashed,
					hostMap, paraMap);
			array[n] = 0;
			fill(array, n + 1, stop, numhost, nodes, weightsolid, weightdashed,
					hostMap, paraMap);
		}
	}

	/*
	 * Calculates the number of violations this coloring gives rise to
	 */

	public static int numViolations(int[] coloring, int numhost,
			InputNode[] nodes, int[][] weightsolid, int[][] weightdashed,
			HashMap<InputHostNode, Integer> hostMap,
			HashMap<InputParasiteNode, Integer> paraMap) {
		int res = 0;
		InputHostNode hostNode;
		InputParasiteNode paraNode;
		for (int i = 0; i < numhost; i++) {
			hostNode = (InputHostNode) nodes[i];
			if (hostMap.containsKey(hostNode)) {
				for (int j = numhost; j < nodes.length; j++) {
					paraNode = (InputParasiteNode) nodes[j];
					if (coloring[i] == coloring[j]) {
						if (weightdashed[hostMap.get(hostNode)][paraMap
								.get(paraNode)] > 0) {
							res++;
						}
					} else {
						if (weightsolid[hostMap.get(hostNode)][paraMap
								.get(paraNode)] > 0) {
							res++;
						}
					}
				}
			}
		}
		return res;
	}

	/**
	 * swap the order of the nodes children
	 * 
	 * @param node
	 */
	private static void swapChildOrder(InputNode node) {
		if (node.getDirectChildrenCount() >= 2) {
			IPositionNode child0 = node.getChildren().get(0);
			IPositionNode child1 = node.getChildren().get(1);
			List<IPositionNode> children = new ArrayList<IPositionNode>();
			children.add(child1);
			children.add(child0);
			node.setChildren(children);
		}
	}

	/**
	 * @param firstHostLeaf
	 * @param firstParasiteLeaf
	 * @param secondHostLeaf
	 * @param secondParasiteLeaf
	 * @param arrangement
	 * @return true if there is a crossing between the associations, otherwise
	 *         false
	 */
	private static boolean isCrossed(InputParasiteNode firstParasiteLeaf,
			InputHostNode firstHostLeaf, InputParasiteNode secondParasiteLeaf,
			InputHostNode secondHostLeaf, ITreeArrangement arrangement) {
		((InputHostNode) NodeUtils.getRootNode(firstHostLeaf)).renumberNodes(0);
		((InputParasiteNode) NodeUtils.getRootNode(firstParasiteLeaf))
				.renumberNodes(0);
		return (((secondHostLeaf.getNumber() - firstHostLeaf.getNumber()) * (secondParasiteLeaf
				.getNumber() - firstParasiteLeaf.getNumber()))
				* ((arrangement
						.isTreeFlipped(InputViewerConstants.TREE_SELECTION_HOST) != arrangement
						.isTreeFlipped(InputViewerConstants.TREE_SELECTION_PARASITE)) ? 1
						: -1) < 0);
	}

}
