/**
 * 
 */
package com.wieseke.cptk.reconstruction.solver;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import com.wieseke.cptk.common.api.IHostNode;
import com.wieseke.cptk.common.api.IParasiteNode;
import com.wieseke.cptk.common.constants.CostsConstants;
import com.wieseke.cptk.common.constants.EventConstants;
import com.wieseke.cptk.common.constants.OptionsConstants;
import com.wieseke.cptk.common.log.CophylogenyLogger;
import com.wieseke.cptk.common.util.CombinationIterator;
import com.wieseke.cptk.common.util.IntegerArrayList;
import com.wieseke.cptk.reconstruction.constants.BaseSolverConstants;
import com.wieseke.cptk.reconstruction.dao.EventType;
import com.wieseke.cptk.reconstruction.dao.EventTypeList;
import com.wieseke.cptk.reconstruction.dao.ReconstructionChildMapping;
import com.wieseke.cptk.reconstruction.dao.ReconstructionCophylogeny;
import com.wieseke.cptk.reconstruction.dao.ReconstructionElement;
import com.wieseke.cptk.reconstruction.dao.ReconstructionHostNode;
import com.wieseke.cptk.reconstruction.dao.ReconstructionInfo;
import com.wieseke.cptk.reconstruction.dao.ReconstructionParasiteNode;
import com.wieseke.cptk.reconstruction.dao.StaticCostTable;
import com.wieseke.cptk.reconstruction.dao.SubReconstructionCost;
import com.wieseke.cptk.reconstruction.util.m2blist.M2BList;
import com.wieseke.cptk.reconstruction.util.m2blist.M2BObject;
import com.wieseke.cptk.reconstruction.util.m2blist.M2BPair;
import com.wieseke.cptk.reconstruction.util.m2blist.M2BSingleValObject;

/**
 * @author Nicolas Wieseke
 * 
 */
public class BaseSolver {

	public void calculateMapping(ReconstructionCophylogeny cophylogeny,
			StaticCostTable staticCostTable) {

		// the parasite nodes will be processed layer by layer from bottom to
		// top
		try {
			for (int level = cophylogeny.getParasiteFirsts().size() - 1; level >= 0; level--) {
				ExecutorService threadExecutor = Executors
						.newFixedThreadPool(BaseSolverConstants.NUMBER_OF_THREADS);
				ReconstructionParasiteNode parasiteNode = cophylogeny
						.getParasiteFirsts().get(level);
				// all parasite node of one layer will be processed
				while (parasiteNode != null) {
					ReconstructionCalculator calculator = new ReconstructionCalculator();
					calculator.init(parasiteNode, cophylogeny, staticCostTable);
					threadExecutor.execute(calculator);
					parasiteNode = parasiteNode.getNext();
				}
				threadExecutor.shutdown();
				while (!threadExecutor.awaitTermination(1000,
						TimeUnit.MILLISECONDS)) {
				}
			}
		} catch (InterruptedException e) {
			CophylogenyLogger.logError(e);
		}
	}

	/**
	 * @author Nicolas Wieseke
	 * 
	 */
	/**
	 * @author Nicolas Wieseke
	 * 
	 */
	@SuppressWarnings("unused")
	public class ReconstructionCalculator implements Runnable {

		/**
		 * the parasite node
		 */
		private ReconstructionParasiteNode parasiteNode;

		/**
		 * the reconstruction cophylogeny
		 */
		private ReconstructionCophylogeny cophylogeny;

		/**
		 * the static cost table
		 */
		private StaticCostTable staticCostTable;

		private String sortingOption = OptionsConstants.SORTING_1_ID;
		private String hostswitchOption = OptionsConstants.HOSTSWITCH_1_ID;
		private String duplicationOption = OptionsConstants.DUPLICATION_0_ID;
		private String takeoffOption = OptionsConstants.TAKEOFF_0_ID;
		private String rankOption = OptionsConstants.RANK_0_ID;
		private String rootMappingOption = OptionsConstants.ROOT_MAPPING_0_ID;
		private String probabilityCostOption = OptionsConstants.PROBABILITY_COSTS_0_ID;
		private String fullHostswitchOption = OptionsConstants.FULL_HOSTSWITCH_0_ID;
		private String leafSpeciacionCostOption = OptionsConstants.LEAF_SPECIACION_COST_0_ID;
		private String checkChronologyOption = OptionsConstants.CHECK_CHRONOLOGY_0_ID;
		private String calcBestCostOption = OptionsConstants.AUTOMATIC_COSTS_0_ID;

		public BigDecimal cospeciationCost = CostsConstants.COSPECIATION_STDCOST;
		public BigDecimal sortingCost = CostsConstants.SORTING_STDCOST;
		public BigDecimal duplicationCost = CostsConstants.DUPLICATION_STDCOST;
		public BigDecimal hostswitchCost = CostsConstants.HOSTSWITCH_STDCOST;

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Runnable#run()
		 */
		@SuppressWarnings("unchecked")
		public void run() {
			try {
				this.parasiteNode.init(cophylogeny.getHostNodes().length);
				for (int hostIndex = 0; hostIndex < cophylogeny.getHostNodes().length; hostIndex++) {
					this.parasiteNode.getReconstructionInfo()[hostIndex] = new ReconstructionInfo(
							this.parasiteNode.getNumber());
				}
				// the node is a leaf node
				if (this.parasiteNode.isLeaf()) {

					// for each valid mapping costs are set to 0
					if (this.parasiteNode.getAssociations() != null) {
						for (ReconstructionHostNode hostNode : this.parasiteNode
								.getAssociations()) {
							if (checkTimezone(this.parasiteNode, hostNode) == 0) {
								ReconstructionInfo reconstructionInfo = new ReconstructionInfo(
										this.parasiteNode.getNumber(), hostNode
												.getNumber(), BigDecimal.ZERO);
								ReconstructionElement reconstruction = new ReconstructionElement(
										this.parasiteNode.getNumber(), hostNode
												.getNumber(), 0);
								reconstruction.setCosts(BigDecimal.ZERO);
								reconstruction.getDerivedEvents().add(
										new EventType(0, 0, 0, 0));
								reconstructionInfo.getReconstructions().add(
										reconstruction);
								this.parasiteNode.setReconstructionInfo(
										hostNode.getNumber(),
										reconstructionInfo);
							}
						}
					}
				}
				// the node is an internal node
				else {
					ReconstructionHostNode[] hostNodes = getValidHostNodes();

					for (ReconstructionHostNode hostNode : hostNodes) {

						// if the mapping is valid
						if (checkTimezone(this.parasiteNode, hostNode) == 0) {

							// initializing stuff for the minimal reconstruction
							BigDecimal minimalCosts = null;
							EventTypeList minimalEventTypesSingle = new EventTypeList();
							IntegerArrayList minimalMappingPositions = new IntegerArrayList();

							// calculate the an array for all reconstruction
							// costs
							// for mapping a parasite child to the different
							// parts
							// in the host tree
							SubReconstructionCost[][] subReconstructionCosts = calculateSubCosts(hostNode);

							// calculate list of subReconstructionCosts indices
							// for
							// each parasite child where the mapping has non
							// infinity costs
							List<Integer>[] possibleChildDomainMappings = computeValidMappingPositions(
									hostNode, subReconstructionCosts);

							// create object containing the possible
							// combinations
							CombinationIterator<Integer> childDomainMappingCombinations = new CombinationIterator<Integer>(
									possibleChildDomainMappings, Integer.class);

							// calculate the reconstruction for each combination
							while (childDomainMappingCombinations.hasNext()) {
								Integer[] childDomaiMapping = childDomainMappingCombinations
										.next();
								// check if all children do host switches and
								// break
								// if this is not permitted

								// compute the events that occurred
								EventType singleEvents = null;
								List<EventType> singleEventsList = null;
								singleEventsList = computeMinimalEventsList(
										childDomaiMapping, hostNode);
								if (singleEventsList != null
										&& singleEventsList.size() > 0)
									singleEvents = singleEventsList.get(0);

								// singleEvents = computeMinimalEvents(
								// childDomaiMapping, hostNode);
								// singleEventsList = new
								// ArrayList<EventType>();
								// singleEventsList.add(singleEvents);

								if (singleEvents != null) {
									// compute the event costs
									BigDecimal singleCosts = computeEventCosts(singleEvents);

									// sum up the total costs
									BigDecimal totalCosts = singleCosts
											.add(BigDecimal.ZERO);
									for (int parasiteChildIndex = 0; parasiteChildIndex < this.parasiteNode
											.getDirectChildrenCount(); parasiteChildIndex++) {
										if (childDomaiMapping[parasiteChildIndex] != null) {
											if (subReconstructionCosts[childDomaiMapping[parasiteChildIndex]][parasiteChildIndex].costs == null) {
												totalCosts = null;
											} else {
												totalCosts = totalCosts
														.add(subReconstructionCosts[childDomaiMapping[parasiteChildIndex]][parasiteChildIndex].costs);
											}
										}
									}

									// compare minimal and current total costs
									if (totalCosts != null) {
										if (minimalCosts == null
												|| minimalCosts
														.compareTo(totalCosts) > 0) {
											minimalCosts = totalCosts;
											minimalEventTypesSingle.clear();
											minimalMappingPositions.clear();
											for (EventType events : singleEventsList) {
												minimalEventTypesSingle
														.add(events);
												minimalMappingPositions
														.add(childDomaiMapping);
											}
										} else if (minimalCosts
												.compareTo(totalCosts) == 0) {
											for (EventType events : singleEventsList) {
												minimalEventTypesSingle
														.add(events);
												minimalMappingPositions
														.add(childDomaiMapping);
											}
										}
									}
								}
							}

							// after checking all possible combinations use the
							// minimal values for the parasite
							ReconstructionInfo reconstructionInfo = new ReconstructionInfo(
									this.parasiteNode.getNumber(), hostNode
											.getNumber(), minimalCosts);
							this.parasiteNode.getReconstructionInfo()[hostNode
									.getNumber()] = reconstructionInfo;
							for (int index = 0; index < minimalEventTypesSingle
									.size(); index++) {

								// compute the combinations for all possible
								// child mappings
								List<Integer>[] possibleChildNodeMappings = new ArrayList[this.parasiteNode
										.getDirectChildrenCount()];
								for (int parasiteChildIndex = 0; parasiteChildIndex < this.parasiteNode
										.getDirectChildrenCount(); parasiteChildIndex++) {
									possibleChildNodeMappings[parasiteChildIndex] = getChildMappingPartnerIndex(
											subReconstructionCosts,
											minimalMappingPositions.get(index),
											parasiteChildIndex);
								}
								CombinationIterator<Integer> childNodeMappingCombinations = new CombinationIterator<Integer>(
										possibleChildNodeMappings,
										Integer.class);

								// calculate the reconstructions for each child
								// mapping combination
								while (childNodeMappingCombinations.hasNext()) {
									Integer[] childNodeMapping = childNodeMappingCombinations
											.next();

									ReconstructionElement currentReconstruction = new ReconstructionElement(
											this.parasiteNode.getNumber(),
											hostNode.getNumber(),
											this.parasiteNode
													.getDirectChildrenCount());
									currentReconstruction
											.setCosts(minimalCosts);

									EventType directEvents = minimalEventTypesSingle
											.get(index).copy();
									EventTypeList[] derivedEventsArray = new EventTypeList[this.parasiteNode
											.getDirectChildrenCount()];

									for (int parasiteChildIndex = 0; parasiteChildIndex < this.parasiteNode
											.getDirectChildrenCount(); parasiteChildIndex++) {
										derivedEventsArray[parasiteChildIndex] = new EventTypeList();
										if (childNodeMapping[parasiteChildIndex] != null) {

											ReconstructionInfo childReconstructionInfo = this.parasiteNode
													.getChildren().get(
															parasiteChildIndex)
													.getReconstructionInfo()[childNodeMapping[parasiteChildIndex]];

											for (ReconstructionElement childReconstruction : childReconstructionInfo
													.getReconstructions()) {
												for (EventType childDerivedEvents : childReconstruction
														.getDerivedEvents()) {
													if (!derivedEventsArray[parasiteChildIndex]
															.contains(childDerivedEvents)) {
														derivedEventsArray[parasiteChildIndex]
																.add(childDerivedEvents);
													}
												}
											}

											directEvents = directEvents
													.add(new EventType(
															0,
															this.staticCostTable.costs[hostNode
																	.getNumber()][childNodeMapping[parasiteChildIndex]]
																	.getSortings(),
															0, 0));

											currentReconstruction
													.getChildReconstructionInfos()[parasiteChildIndex] = childReconstructionInfo;

										}
										if (childNodeMapping[parasiteChildIndex] == null) {
											currentReconstruction
													.getChildMappings()[parasiteChildIndex] = new ReconstructionChildMapping();
										} else {
											int childMapping = childNodeMapping[parasiteChildIndex];
											List<Integer> takeof = staticCostTable.costs[hostNode
													.getNumber()][childMapping]
													.getTakeof();
											List<Integer> landing = staticCostTable.costs[hostNode
													.getNumber()][childMapping]
													.getLanding();
											boolean hostswitch = staticCostTable.costs[hostNode
													.getNumber()][childMapping]
													.isHostswitch();
											currentReconstruction
													.getChildMappings()[parasiteChildIndex] = new ReconstructionChildMapping(
													childMapping, takeof,
													landing, hostswitch);
										}
									}
									currentReconstruction
											.setDirectEvents(directEvents);

									List<EventType> derivedEventsList = new ArrayList<EventType>();
									CombinationIterator<EventType> childEventTypeCombinations = new CombinationIterator<EventType>(
											derivedEventsArray, EventType.class);

									// calculate the derived events for each
									// child's derived events combination
									while (childEventTypeCombinations.hasNext()) {
										EventType[] childDerivedEvents = childEventTypeCombinations
												.next();

										EventType derivedEvents = new EventType(
												0, 0, 0, 0).add(directEvents);
										for (int parasiteChildIndex = 0; parasiteChildIndex < this.parasiteNode
												.getDirectChildrenCount(); parasiteChildIndex++) {
											if (childDerivedEvents[parasiteChildIndex] != null) {
												derivedEvents = derivedEvents
														.add(childDerivedEvents[parasiteChildIndex]);
											}
										}
										// Option 1: only use paretofront
										// TODO PARETOFRONT
										boolean insertEvents = true;
										Object[] derivedEventsListArray = derivedEventsList
												.toArray();

										for (Object itemObj : derivedEventsListArray) {
											EventType item = (EventType) itemObj;
											if (item.getCospeciations() <= derivedEvents
													.getCospeciations()
													&& item.getSortings() <= derivedEvents
															.getSortings()
													&& item.getDuplications() <= derivedEvents
															.getDuplications()
													&& item.getHostswitches() <= derivedEvents
															.getHostswitches()) {
												insertEvents = false;
												break;
											}
											if (item.getCospeciations() >= derivedEvents
													.getCospeciations()
													&& item.getSortings() >= derivedEvents
															.getSortings()
													&& item.getDuplications() >= derivedEvents
															.getDuplications()
													&& item.getHostswitches() >= derivedEvents
															.getHostswitches()) {
												derivedEventsList.remove(item);
											}
										}
										if (insertEvents) {
											derivedEventsList
													.add(derivedEvents);
										}

										// Option 2: use all
										// if (!derivedEventsList
										// .contains(derivedEvents)) {
										// derivedEventsList
										// .add(derivedEvents);
										// }

									}

									currentReconstruction
											.setDerivedEvents(derivedEventsList);

									reconstructionInfo.getReconstructions()
											.add(currentReconstruction);

								}
							}

							if (checkChronologyOption
									.equals(OptionsConstants.CHECK_CHRONOLOGY_1_ID)) {
								this.parasiteNode.chronology[hostNode
										.getNumber()] = checkChronology(hostNode);
							}
						}
					}

				}
			} catch (Exception e) {
				CophylogenyLogger.logError(e);
			}
		}

		/**
		 * @return an array containing all valid host nodes as mapping partner
		 *         for the parasite node
		 */
		private ReconstructionHostNode[] getValidHostNodes() {
			ReconstructionHostNode[] hostNodes;
			if (this.parasiteNode.getAssociations() == null
					|| this.parasiteNode.getAssociations().size() == 0) {
				hostNodes = this.cophylogeny.getHostNodes();
			} else {
				hostNodes = new ReconstructionHostNode[this.parasiteNode
						.getAssociations().size()];
				for (int index = 0; index < this.parasiteNode.getAssociations()
						.size(); index++) {
					hostNodes[index] = this.parasiteNode.getAssociations().get(
							index);
				}
			}
			return hostNodes;
		}

		/**
		 * @param mappingPosition
		 * @param hostNode
		 * @return EventType object containing the number of events that
		 *         occurred
		 */
		private EventType computeMinimalEvents(Integer[] mappingPosition,
				ReconstructionHostNode hostNode) {

			int usedChildren = 0;
			int[] usedPositions = new int[hostNode.getDirectChildrenCount() + 2];
			for (int index = 0; index < mappingPosition.length; index++) {
				if (mappingPosition[index] != null) {
					usedPositions[mappingPosition[index]]++;
					usedChildren++;
				}
			}
			int duplications = 0;
			int cospeciations = 0;
			int sortings = 0;
			int hostswitches = usedPositions[usedPositions.length - 1];
			// return null if no children where used or full host switches are
			// not permitted and all children do host switches
			if (usedChildren == 0
					|| (fullHostswitchOption
							.equals(OptionsConstants.FULL_HOSTSWITCH_0_ID) && usedChildren == hostswitches)) {
				return null;
			}
			// wenn ein Parasitenkind auf den
			// gleichen Hostknoten gesetzt wird, so
			// muss die
			// Speziation des Parasiten vor der des
			// Hosts stattgefunden haben
			// es gibt somit keine Cospeziation,
			// sondern nur Duplikationen für alle
			// Kinder,
			// die keinen Hostswitch durchführen
			if (usedPositions[0] > 0 || !checkPossibleCospeciation(hostNode)) {
				for (int j = 0; j < usedPositions.length - 1; j++) {
					if (usedPositions[j] > 0) {
						duplications += usedPositions[j];
						if (j > 0) {
							sortings += usedPositions[j];
						}
					}
				}
				// zuvielgezählte Duplikation
				// abziehen
				if (duplications > 0) {
					duplications--;
				}
			}
			// anderenfalls muss geprüft werden mit
			// wie vielen Cospeziationen das Mapping
			// am günstigsten wird
			else {
				// childSum = Summe der benutzten
				// Kinder, die keine Hostswitches
				// durchführen. Die Summe bestimmt
				// die maximale Anzahl an Sortings
				int childSum = 0;
				ArrayList<M2BObject> m2bUsedPositions = new ArrayList<M2BObject>();
				// alle Mappings der Parasitenkinder
				// in die Hostkinder werden
				// durchsucht
				for (int j = 1; j < usedPositions.length - 1; j++) {
					childSum += usedPositions[j];
					m2bUsedPositions.add(new M2BSingleValObject(
							usedPositions[j]));
				}
				M2BList m2blist = new M2BList(m2bUsedPositions);

				M2BPair m2bpair = m2blist.getPair();
				int maxCospecNumber = 0;
				while (m2bpair.first != null && m2bpair.second != null) {
					maxCospecNumber++;
					m2bpair = m2blist.getPair();
				}

				BigDecimal eventcosts = null;
				int tempCospeciations = 0;
				int tempDuplications = 0;
				int tempSortings = 0;
				int tempSortings2 = 0;
				// für den Fall, dass der Knoten nur
				// ein Pseudoknoten ist, der
				// eingefügt wurde, weil ein
				// Parasitenblatt auf mehrere
				// Hostblätter abgebildet wurde
				// dann werden Cospeziationen und
				// Duplikationen nicht mitgerechnet
				// dieser Fall ist nur in der
				// esle-Verzweigung von Interesse,
				// da ein Pseudoknoten kann nicht
				// auf einem Hostknoten sitzen kann,
				// auf dem mehr als nur eines seiner
				// Kinder sitzt, denn dies müsste
				// ein Hostblatt sein, und dem ist
				// nur eines der Parasitenkinder
				// zugeordnet.
				// in einem solchen Fall müssen alle
				// anderen Kinder Hostswitches
				// machen und die werden sowieso
				// mitgezählt
				if (leafSpeciacionCostOption
						.equals(OptionsConstants.LEAF_SPECIACION_COST_0_ID)
						&& checkLeafPseudoSpeciation()) {
					for (int j = 0; j < m2bUsedPositions.size(); j++) {
						tempSortings += m2bUsedPositions.get(j).getKey();
						if (m2bUsedPositions.get(j).getKey() > 0)
							tempSortings2++;
					}
					// tempSortings2 bestimmt die
					// Anzahl der Sortings, wenn die
					// Duplikationen unter den
					// Vaterknoten gezogen werden
					// dies ist aber nur möglich,
					// wenn oberhalb des Vaterknoten
					// überhaupt was passiert
					// (Hostswitches oder zumindest
					// Verzweigungen in verschiedene
					// Äste)
					if (hostswitches == 0 && tempSortings2 == 1)
						tempSortings2 = tempSortings;

					BigDecimal tempCosts1 = BigDecimal.valueOf(
							tempCospeciations).multiply(cospeciationCost).add(
							BigDecimal.valueOf(tempDuplications).multiply(
									duplicationCost)).add(
							BigDecimal.valueOf(tempSortings).multiply(
									sortingCost));
					BigDecimal tempCosts2 = BigDecimal.valueOf(
							tempCospeciations).multiply(cospeciationCost).add(
							BigDecimal.valueOf(tempDuplications).multiply(
									duplicationCost)).add(
							BigDecimal.valueOf(tempSortings2).multiply(
									sortingCost));
					if (eventcosts == null
							|| tempCosts1.compareTo(eventcosts) < 0) {
						eventcosts = tempCosts1;
						cospeciations = tempCospeciations;
						duplications = tempDuplications;
						sortings = tempSortings;
					}
					if (eventcosts == null
							|| tempCosts2.compareTo(eventcosts) < 0) {
						eventcosts = tempCosts2;
						cospeciations = tempCospeciations;
						duplications = tempDuplications;
						sortings = tempSortings2;
					}
				}
				// kein Pseudoknoten
				else {
					for (int cospec = 0; cospec <= maxCospecNumber; cospec++) {
						tempCospeciations = cospec;
						tempDuplications = Math.max(0, childSum - 1 - cospec);
						// 2 verschidene Varianten
						// der Berechnung
						// zusätzlicher Sortings
						// 1. Duplikationen können
						// nicht unter den
						// Vaterknoten verschoben
						// werden.
						// Das heisst, dass alle
						// Kinder die nicht an
						// Cospeziationen teilnehmen
						// zusätzlice Sortings
						// durchführen müssen
						tempSortings = Math.max(0, childSum - (cospec * 2));
						// 2. Duplikationen können
						// unter den Vaterknoten
						// verschoben werden
						// dann mus nur noch ein
						// Sorting für jeden Ast,
						// der keine Cospeziation
						// abbekommt, durchgeführt
						// werden
						tempSortings2 = 0;
						for (int j = cospec * 2; j < m2bUsedPositions.size(); j++) {
							tempSortings2++;
						}
						// tempSortings2 bestimmt
						// die Anzahl der Sortings,
						// wenn die Duplikationen
						// unter den Vaterknoten
						// gezogen werden
						// dies ist aber nur
						// möglich, wenn oberhalb
						// des Vaterknoten überhaupt
						// was passiert
						// (Hostswitches oder
						// zumindest Verzweigungen
						// in verschiedene Äste)
						if (hostswitches == 0 && maxCospecNumber == 0)
							tempSortings2 = tempSortings;

						BigDecimal tempCosts1 = BigDecimal.valueOf(
								tempCospeciations).multiply(cospeciationCost)
								.add(
										BigDecimal.valueOf(tempDuplications)
												.multiply(duplicationCost))
								.add(
										BigDecimal.valueOf(tempSortings)
												.multiply(sortingCost));
						BigDecimal tempCosts2 = BigDecimal.valueOf(
								tempCospeciations).multiply(cospeciationCost)
								.add(
										BigDecimal.valueOf(tempDuplications)
												.multiply(duplicationCost))
								.add(
										BigDecimal.valueOf(tempSortings2)
												.multiply(sortingCost));
						if (eventcosts == null
								|| tempCosts1.compareTo(eventcosts) < 0) {
							eventcosts = tempCosts1;
							cospeciations = tempCospeciations;
							duplications = tempDuplications;
							sortings = tempSortings;
						}
						if (eventcosts == null
								|| tempCosts2.compareTo(eventcosts) < 0) {
							eventcosts = tempCosts2;
							cospeciations = tempCospeciations;
							duplications = tempDuplications;
							sortings = tempSortings2;
						}
					}
				}
			}
			if (this.duplicationOption
					.equals(OptionsConstants.DUPLICATION_0_ID))
				duplications += hostswitches;
			return new EventType(cospeciations, sortings, duplications,
					hostswitches);
		}

		/**
		 * @param mappingPosition
		 * @param hostNode
		 * @return EventType object containing the number of events that
		 *         occurred
		 */
		private List<EventType> computeMinimalEventsList(
				Integer[] mappingPosition, ReconstructionHostNode hostNode) {

			List<EventType> result = new ArrayList<EventType>();
			int usedChildren = 0;
			int[] usedPositions = new int[hostNode.getDirectChildrenCount() + 2];
			for (int index = 0; index < mappingPosition.length; index++) {
				if (mappingPosition[index] != null) {
					usedPositions[mappingPosition[index]]++;
					usedChildren++;
				}
			}
			List<Integer> duplications = new ArrayList<Integer>();
			List<Integer> cospeciations = new ArrayList<Integer>();
			List<Integer> sortings = new ArrayList<Integer>();
			int hostswitches = usedPositions[usedPositions.length - 1];
			// return null if no children where used or full host switches are
			// not permitted and all children do host switches
			if (usedChildren == 0
					|| (fullHostswitchOption
							.equals(OptionsConstants.FULL_HOSTSWITCH_0_ID) && usedChildren == hostswitches)) {
				return null;
			}
			// wenn ein Parasitenkind auf den
			// gleichen Hostknoten gesetzt wird, so
			// muss die
			// Speziation des Parasiten vor der des
			// Hosts stattgefunden haben
			// es gibt somit keine Cospeziation,
			// sondern nur Duplikationen für alle
			// Kinder,
			// die keinen Hostswitch durchführen
			if (usedPositions[0] > 0 || !checkPossibleCospeciation(hostNode)) {
				int tempDuplications = 0;
				int tempSortings = 0;
				for (int j = 0; j < usedPositions.length - 1; j++) {
					if (usedPositions[j] > 0) {
						tempDuplications += usedPositions[j];
						if (j > 0) {
							tempSortings += usedPositions[j];
						}
					}
				}
				// zuvielgezählte Duplikation
				// abziehen
				if (tempDuplications > 0) {
					tempDuplications--;
				}
				cospeciations.add(0);
				duplications.add(tempDuplications);
				sortings.add(tempSortings);
			}
			// anderenfalls muss geprüft werden mit
			// wie vielen Cospeziationen das Mapping
			// am günstigsten wird
			else {
				// childSum = Summe der benutzten
				// Kinder, die keine Hostswitches
				// durchführen. Die Summe bestimmt
				// die maximale Anzahl an Sortings
				int childSum = 0;
				ArrayList<M2BObject> m2bUsedPositions = new ArrayList<M2BObject>();
				// alle Mappings der Parasitenkinder
				// in die Hostkinder werden
				// durchsucht
				for (int j = 1; j < usedPositions.length - 1; j++) {
					childSum += usedPositions[j];
					m2bUsedPositions.add(new M2BSingleValObject(
							usedPositions[j]));
				}
				M2BList m2blist = new M2BList(m2bUsedPositions);

				M2BPair m2bpair = m2blist.getPair();
				int maxCospecNumber = 0;
				while (m2bpair.first != null && m2bpair.second != null) {
					maxCospecNumber++;
					m2bpair = m2blist.getPair();
				}

				BigDecimal eventcosts = null;
				int tempCospeciations = 0;
				int tempDuplications = 0;
				int tempSortings = 0;
				int tempSortings2 = 0;
				// für den Fall, dass der Knoten nur
				// ein Pseudoknoten ist, der
				// eingefügt wurde, weil ein
				// Parasitenblatt auf mehrere
				// Hostblätter abgebildet wurde
				// dann werden Cospeziationen und
				// Duplikationen nicht mitgerechnet
				// dieser Fall ist nur in der
				// esle-Verzweigung von Interesse,
				// da ein Pseudoknoten kann nicht
				// auf einem Hostknoten sitzen kann,
				// auf dem mehr als nur eines seiner
				// Kinder sitzt, denn dies müsste
				// ein Hostblatt sein, und dem ist
				// nur eines der Parasitenkinder
				// zugeordnet.
				// in einem solchen Fall müssen alle
				// anderen Kinder Hostswitches
				// machen und die werden sowieso
				// mitgezählt
				if (leafSpeciacionCostOption
						.equals(OptionsConstants.LEAF_SPECIACION_COST_0_ID)
						&& checkLeafPseudoSpeciation()) {
					for (int j = 0; j < m2bUsedPositions.size(); j++) {
						tempSortings += m2bUsedPositions.get(j).getKey();
						if (m2bUsedPositions.get(j).getKey() > 0)
							tempSortings2++;
					}
					// tempSortings2 bestimmt die
					// Anzahl der Sortings, wenn die
					// Duplikationen unter den
					// Vaterknoten gezogen werden
					// dies ist aber nur möglich,
					// wenn oberhalb des Vaterknoten
					// überhaupt was passiert
					// (Hostswitches oder zumindest
					// Verzweigungen in verschiedene
					// Äste)
					if (hostswitches == 0 && tempSortings2 == 1)
						tempSortings2 = tempSortings;

					BigDecimal tempCosts1 = BigDecimal.valueOf(
							tempCospeciations).multiply(cospeciationCost).add(
							BigDecimal.valueOf(tempDuplications).multiply(
									duplicationCost)).add(
							BigDecimal.valueOf(tempSortings).multiply(
									sortingCost));
					BigDecimal tempCosts2 = BigDecimal.valueOf(
							tempCospeciations).multiply(cospeciationCost).add(
							BigDecimal.valueOf(tempDuplications).multiply(
									duplicationCost)).add(
							BigDecimal.valueOf(tempSortings2).multiply(
									sortingCost));
					if (eventcosts == null
							|| tempCosts1.compareTo(eventcosts) < 0) {
						eventcosts = tempCosts1;
						cospeciations.clear();
						duplications.clear();
						sortings.clear();
						cospeciations.add(tempCospeciations);
						duplications.add(tempDuplications);
						sortings.add(tempSortings);
					} else if (eventcosts == null
							|| tempCosts1.compareTo(eventcosts) == 0) {
						eventcosts = tempCosts1;
						cospeciations.add(tempCospeciations);
						duplications.add(tempDuplications);
						sortings.add(tempSortings);
					}
					if (eventcosts == null
							|| tempCosts2.compareTo(eventcosts) < 0) {
						eventcosts = tempCosts2;
						cospeciations.clear();
						duplications.clear();
						sortings.clear();
						cospeciations.add(tempCospeciations);
						duplications.add(tempDuplications);
						sortings.add(tempSortings2);
					} else if (eventcosts == null
							|| tempCosts2.compareTo(eventcosts) == 0) {
						eventcosts = tempCosts2;
						cospeciations.add(tempCospeciations);
						duplications.add(tempDuplications);
						sortings.add(tempSortings2);
					}
				}
				// kein Pseudoknoten
				else {
					for (int cospec = 0; cospec <= maxCospecNumber; cospec++) {
						tempCospeciations = cospec;
						tempDuplications = Math.max(0, childSum - 1 - cospec);
						// 2 verschidene Varianten
						// der Berechnung
						// zusätzlicher Sortings
						// 1. Duplikationen können
						// nicht unter den
						// Vaterknoten verschoben
						// werden.
						// Das heisst, dass alle
						// Kinder die nicht an
						// Cospeziationen teilnehmen
						// zusätzlice Sortings
						// durchführen müssen
						tempSortings = Math.max(0, childSum - (cospec * 2));
						// 2. Duplikationen können
						// unter den Vaterknoten
						// verschoben werden
						// dann mus nur noch ein
						// Sorting für jeden Ast,
						// der keine Cospeziation
						// abbekommt, durchgeführt
						// werden
						tempSortings2 = 0;
						for (int j = cospec * 2; j < m2bUsedPositions.size(); j++) {
							tempSortings2++;
						}
						// tempSortings2 bestimmt
						// die Anzahl der Sortings,
						// wenn die Duplikationen
						// unter den Vaterknoten
						// gezogen werden
						// dies ist aber nur
						// möglich, wenn oberhalb
						// des Vaterknoten überhaupt
						// was passiert
						// (Hostswitches oder
						// zumindest Verzweigungen
						// in verschiedene Äste)
						if (hostswitches == 0 && maxCospecNumber == 0)
							tempSortings2 = tempSortings;

						BigDecimal tempCosts1 = BigDecimal.valueOf(
								tempCospeciations).multiply(cospeciationCost)
								.add(
										BigDecimal.valueOf(tempDuplications)
												.multiply(duplicationCost))
								.add(
										BigDecimal.valueOf(tempSortings)
												.multiply(sortingCost));
						BigDecimal tempCosts2 = BigDecimal.valueOf(
								tempCospeciations).multiply(cospeciationCost)
								.add(
										BigDecimal.valueOf(tempDuplications)
												.multiply(duplicationCost))
								.add(
										BigDecimal.valueOf(tempSortings2)
												.multiply(sortingCost));
						if (eventcosts == null
								|| tempCosts1.compareTo(eventcosts) < 0) {
							eventcosts = tempCosts1;
							cospeciations.clear();
							duplications.clear();
							sortings.clear();
							cospeciations.add(tempCospeciations);
							duplications.add(tempDuplications);
							sortings.add(tempSortings);
						} else if (eventcosts == null
								|| tempCosts1.compareTo(eventcosts) == 0) {
							eventcosts = tempCosts1;
							cospeciations.add(tempCospeciations);
							duplications.add(tempDuplications);
							sortings.add(tempSortings);
						}
						if (eventcosts == null
								|| tempCosts2.compareTo(eventcosts) < 0) {
							eventcosts = tempCosts2;
							cospeciations.clear();
							duplications.clear();
							sortings.clear();
							cospeciations.add(tempCospeciations);
							duplications.add(tempDuplications);
							sortings.add(tempSortings2);
						} else if (eventcosts == null
								|| tempCosts2.compareTo(eventcosts) == 0) {
							eventcosts = tempCosts2;
							cospeciations.add(tempCospeciations);
							duplications.add(tempDuplications);
							sortings.add(tempSortings2);
						}
					}
				}
			}
			for (int i = 0; i < cospeciations.size(); i++) {
				int tempDuplications = duplications.get(i);
				if (this.duplicationOption
						.equals(OptionsConstants.DUPLICATION_0_ID))
					tempDuplications += hostswitches;
				EventType events = new EventType(cospeciations.get(i)
						.intValue(), sortings.get(i).intValue(),
						tempDuplications, hostswitches);

				// boolean insertEvents = true;
				// for (EventType item : result) {
				// if (item.getCospeciations() <= events.getCospeciations()
				// && item.getSortings() <= events.getSortings()
				// && item.getDuplications() <= events
				// .getDuplications()
				// && item.getHostswitches() <= events
				// .getHostswitches()) {
				// insertEvents = false;
				// break;
				// }
				// if (item.getCospeciations() >= events.getCospeciations()
				// && item.getSortings() >= events.getSortings()
				// && item.getDuplications() >= events
				// .getDuplications()
				// && item.getHostswitches() >= events
				// .getHostswitches()) {
				// result.remove(item);
				// }
				// }
				// if (insertEvents)
				if (!result.contains(events))
					result.add(events);
			}
			return result;
		}

		/**
		 * @param events
		 * @return the sum of the occurred events (excluding the host switch
		 *         costs as they are computed seperately in the static cost
		 *         table)
		 */
		private BigDecimal computeEventCosts(EventType events) {
			return this.cospeciationCost.multiply(
					BigDecimal.valueOf(events.getCospeciations())).add(
					this.sortingCost.multiply(BigDecimal.valueOf(events
							.getSortings()))).add(
					this.duplicationCost.multiply(BigDecimal.valueOf(events
							.getDuplications())));
			// + this.hostswitchCost * events.getHostswitches();
		}

		/**
		 * @param subReconstructionCost
		 * @param mappingPositions
		 * @param parasiteChildIndex
		 * @return a list of possible mapping partner for a parasite child
		 */
		public List<Integer> getChildMappingPartnerIndex(
				SubReconstructionCost[][] subReconstructionCost,
				Integer[] mappingPositions, int parasiteChildIndex) {
			if (mappingPositions[parasiteChildIndex] == null)
				return new ArrayList<Integer>();
			return subReconstructionCost[mappingPositions[parasiteChildIndex]][parasiteChildIndex].hostNodes;
		}

		/**
		 * @param hostNode
		 * @return true if the cronology is valid
		 */
		private boolean checkChronology(ReconstructionHostNode hostNode) {
			// TODO checkChronology during reconstruction
			return false;
		}

		private boolean checkLeafPseudoSpeciation() {
			// TODO checkPossibleCospeciation
			for (ReconstructionParasiteNode child : this.parasiteNode
					.getChildren()) {
				if (!child.getLabel().startsWith(
						this.parasiteNode.getLabel() + "#")) {
					return false;
				}
			}
			return true;
		}

		private boolean checkPossibleCospeciation(
				ReconstructionHostNode hostNode) {
			if (this.rankOption.equals(OptionsConstants.RANK_1_ID)
					&& this.parasiteNode.getRank().getRankTo() < hostNode
							.getRank()) {
				return false;
			}
			return true;
		}

		/**
		 * calculate a lists for each parasite child containing the indices of
		 * subReconstructionCosts with non infinity costs
		 * 
		 * @param hostNode
		 * @param subReconstructionCosts
		 * @return
		 */
		@SuppressWarnings("unchecked")
		private List<Integer>[] computeValidMappingPositions(
				ReconstructionHostNode hostNode,
				SubReconstructionCost[][] subReconstructionCosts) {
			List<Integer>[] validMappingPositions = new ArrayList[this.parasiteNode
					.getDirectChildrenCount()];
			for (int i = 0; i < this.parasiteNode.getDirectChildrenCount(); i++) {
				validMappingPositions[i] = new ArrayList<Integer>();
				for (int j = 0; j < hostNode.getDirectChildrenCount() + 2; j++) {
					if (subReconstructionCosts[j][i].costs != null)
						validMappingPositions[i].add(j);
				}
			}
			return validMappingPositions;
		}

		/**
		 * calculates the best costs for mapping each parasite child to either
		 * the same host node [0], to one of the subtrees starting at one of the
		 * children of the host node [1]-[parasiteNode.getDirectChildrenCount()]
		 * or to somewhere else in the host tree
		 * [parasiteNode.getDirectChildrenCount()+1]
		 * 
		 * @param hostNode
		 * @return
		 */
		private SubReconstructionCost[][] calculateSubCosts(
				ReconstructionHostNode hostNode) {
			SubReconstructionCost[][] subCosts = new SubReconstructionCost[hostNode
					.getDirectChildrenCount() + 2][this.parasiteNode
					.getDirectChildrenCount()];
			for (int i = 0; i < hostNode.getDirectChildrenCount() + 2; i++) {
				for (int j = 0; j < this.parasiteNode.getDirectChildrenCount(); j++) {
					subCosts[i][j] = new SubReconstructionCost();
				}
			}
			for (int parasiteChildIndex = 0; parasiteChildIndex < this.parasiteNode
					.getDirectChildrenCount(); parasiteChildIndex++) {
				ReconstructionParasiteNode parasiteChild = this.parasiteNode
						.getChildren().get(parasiteChildIndex);
				for (int hostMappingIndex = 0; hostMappingIndex < this.cophylogeny
						.getHostNodes().length; hostMappingIndex++) {
					ReconstructionHostNode childMappingPartner = this.cophylogeny
							.getHostNodes()[hostMappingIndex];
					BigDecimal costs = null;
					if (parasiteChild.getReconstructionInfo()[hostMappingIndex]
							.getCosts() != null
							&& this.staticCostTable.costs[hostNode.getNumber()][hostMappingIndex]
									.getCost() != null) {
						costs = parasiteChild.getReconstructionInfo()[hostMappingIndex]
								.getCosts().add(
										this.staticCostTable.costs[hostNode
												.getNumber()][hostMappingIndex]
												.getCost());
					}
					int mappingPosition = getMappingPosition(hostNode,
							childMappingPartner);
					if (costs != null) {
						if (subCosts[mappingPosition][parasiteChildIndex].costs == null
								|| subCosts[mappingPosition][parasiteChildIndex].costs
										.compareTo(costs) > 0) {
							subCosts[mappingPosition][parasiteChildIndex] = new SubReconstructionCost();
							subCosts[mappingPosition][parasiteChildIndex].costs = costs;
							subCosts[mappingPosition][parasiteChildIndex].hostNodes
									.add(hostMappingIndex);
						} else if (subCosts[mappingPosition][parasiteChildIndex].costs
								.compareTo(costs) == 0) {
							subCosts[mappingPosition][parasiteChildIndex].hostNodes
									.add(hostMappingIndex);
						}
					}
				}
			}
			return subCosts;
		}

		private int getMappingPosition(ReconstructionHostNode hostNode,
				ReconstructionHostNode childMappingPartner) {
			if (hostNode.getNumber() == childMappingPartner.getNumber())
				return 0;
			for (int index = 0; index < hostNode.getDirectChildrenCount(); index++) {
				ReconstructionHostNode hostChild = hostNode.getChildren().get(
						index);
				if (childMappingPartner.getNumber() >= hostChild.getNumber()
						&& childMappingPartner.getNumber() <= hostChild
								.getNumber()
								+ hostChild.getChildrenCount())
					return index + 1;
			}
			return hostNode.getDirectChildrenCount() + 1;
		}

		/**
		 * sets the start and end node
		 * 
		 * @param start
		 * @param end
		 */
		public void init(ReconstructionParasiteNode parasiteNode,
				ReconstructionCophylogeny cophylogeny,
				StaticCostTable staticCostTable) {
			this.parasiteNode = parasiteNode;
			this.cophylogeny = cophylogeny;
			this.staticCostTable = staticCostTable;

			if (cophylogeny.getOptions().get(OptionsConstants.SORTING_ID) != null)
				this.sortingOption = cophylogeny.getOptions().get(
						OptionsConstants.SORTING_ID);
			if (cophylogeny.getOptions().get(OptionsConstants.HOSTSWITCH_ID) != null)
				this.hostswitchOption = cophylogeny.getOptions().get(
						OptionsConstants.HOSTSWITCH_ID);
			if (cophylogeny.getOptions().get(OptionsConstants.DUPLICATION_ID) != null)
				this.duplicationOption = cophylogeny.getOptions().get(
						OptionsConstants.DUPLICATION_ID);
			if (cophylogeny.getOptions().get(OptionsConstants.TAKEOFF_ID) != null)
				this.takeoffOption = cophylogeny.getOptions().get(
						OptionsConstants.TAKEOFF_ID);
			if (cophylogeny.getOptions().get(OptionsConstants.RANK_ID) != null)
				this.rankOption = cophylogeny.getOptions().get(
						OptionsConstants.RANK_ID);
			if (cophylogeny.getOptions().get(OptionsConstants.ROOT_MAPPING_ID) != null)
				this.rootMappingOption = cophylogeny.getOptions().get(
						OptionsConstants.ROOT_MAPPING_ID);
			if (cophylogeny.getOptions().get(
					OptionsConstants.PROBABILITY_COSTS_ID) != null)
				this.probabilityCostOption = cophylogeny.getOptions().get(
						OptionsConstants.PROBABILITY_COSTS_ID);
			if (cophylogeny.getOptions().get(
					OptionsConstants.FULL_HOSTSWITCH_ID) != null)
				this.fullHostswitchOption = cophylogeny.getOptions().get(
						OptionsConstants.FULL_HOSTSWITCH_ID);
			if (cophylogeny.getOptions().get(
					OptionsConstants.LEAF_SPECIACION_COST_ID) != null)
				this.leafSpeciacionCostOption = cophylogeny.getOptions().get(
						OptionsConstants.LEAF_SPECIACION_COST_ID);
			if (cophylogeny.getOptions().get(
					OptionsConstants.CHECK_CHRONOLOGY_ID) != null)
				this.checkChronologyOption = cophylogeny.getOptions().get(
						OptionsConstants.CHECK_CHRONOLOGY_ID);
			if (cophylogeny.getOptions().get(
					OptionsConstants.AUTOMATIC_COSTS_ID) != null)
				this.calcBestCostOption = cophylogeny.getOptions().get(
						OptionsConstants.AUTOMATIC_COSTS_ID);

			if (cophylogeny.getCosts().get(EventConstants.COSPECIATION) != null)
				this.cospeciationCost = cophylogeny.getCosts().get(
						EventConstants.COSPECIATION);
			if (cophylogeny.getCosts().get(EventConstants.SORTING) != null)
				this.sortingCost = cophylogeny.getCosts().get(
						EventConstants.SORTING);
			if (cophylogeny.getCosts().get(EventConstants.DUPLICATION) != null)
				this.duplicationCost = cophylogeny.getCosts().get(
						EventConstants.DUPLICATION);
			if (cophylogeny.getCosts().get(EventConstants.HOSTSWITCH) != null)
				this.hostswitchCost = cophylogeny.getCosts().get(
						EventConstants.HOSTSWITCH);
		}

		/**
		 * @return -1 if the time zone of start is before time zone of end, 1 if
		 *         the time zone of start is after time zone of end and 0 if the
		 *         time zones overlap
		 */
		private int checkTimezone(IParasiteNode parasiteNode, IHostNode hostNode) {
			if (this.cophylogeny.getOptions().get(OptionsConstants.RANK_ID)
					.equals(OptionsConstants.RANK_1_ID)) {
				int hostTimezoneFrom = hostNode.getParent() == null ? Integer.MIN_VALUE
						: hostNode.getParent().getRank();
				int hostTimezoneTo = hostNode.getRank();
				if (parasiteNode.getRank().getRankFrom() > hostTimezoneTo)
					return 1;
				if (parasiteNode.getRank().getRankTo() < hostTimezoneFrom)
					return -1;
			}
			return 0;
		}

	}
}
