/**
 * 
 */
package com.wieseke.cptk.reconstruction.util;

import java.math.BigDecimal;
import java.math.RoundingMode;

import com.wieseke.cptk.common.constants.EventConstants;
import com.wieseke.cptk.common.dao.ReconstructionCosts;
import com.wieseke.cptk.reconstruction.constants.BaseSolverConstants;
import com.wieseke.cptk.reconstruction.dao.EventType;

/**
 * @author Nic
 * 
 */
@SuppressWarnings("unused")
public class QualityCalculator {

	/**
	 * computes the quality value for the used costs and the given event types
	 * 
	 * @param costs
	 * @param events
	 * @return
	 */
	public static BigDecimal computeQuality(ReconstructionCosts costs,
			EventType events) {
		return computeQuality1(costs, events);
	}

	/**
	 * computes the quality value for the used costs and the given event types
	 * all events with factors
	 * 
	 * @param costs
	 * @param events
	 * @return
	 */
	private static BigDecimal computeQuality0(ReconstructionCosts costs,
			EventType events) {
		try {
			// standard
			double[] factors = { 0.25d, 0.25d, 0.25d, 0.25d };
			// charleston
			// double[] factors = { 0.13024644897857401d, 0.1860693440598079d,
			// 0.05921079236334392d, 0.6244734145982742d };
			// id
			// double[] factors = { 0.3973264489807361d, 0.1832419792126688d,
			// 0.007144588309055788d, 0.41228698349753934d };
			// gopher-lice
			// double[] factors = { 0.20210767960165565d, 0.19169768253382086d,
			// 0.04906582351695855d, 0.557128814347565d };
			// seabirds-chewinglice
			// double[] factors = { 0.19753173046811756d, 0.18307616978342722d,
			// 0.04046705731633751d, 0.5789250424321177d };
			// double[] factors = { 0.3141688108117421, 0.33244703236961204,
			// 0.16723622494972187, 0.18614793186892398 };
			// pelecan-lice-ml
			// double[] factors = { 0.18699067596070532d, 0.25975568364218604d,
			// 0.06517324615159412d, 0.4880803942455145d };
			// pelecan-lice-mp
			// double[] factors = { 0.23016934687522778d, 0.21401128252714607d,
			// 0.064434446844689942d, 0.4913849237529362d };
			// caryo-micro
			// double[] factors = { 0.0920295995430192d, 0.12152703245427501d,
			// 0.03366702960296283d, 0.752776338399743d };
			// 1a-ficus
			// double[] factors = { 0.338079519052663d, 0.21844638054362736d,
			// 0.06706937653792736d, 0.3764047238657824d };
			// 3-ficus
			// double[] factors = { 0.15612785889585418d, 0.14846741852249484d,
			// 0.164065530854494d, 0.531339191727157d };
			// rodent-lice
			// double[] factors = { 0.2924783456280653d, 0.268727295922778d,
			// 0.06946224040649279d, 0.36933211804266386d };

			BigDecimal reciprocalCostSum = BigDecimal.ZERO;
			BigDecimal probabilities_nn[] = new BigDecimal[costs.values()
					.size()];
			int index = 0;
			for (String eventName : EventConstants.getEvents()) {
				if (costs.get(eventName).compareTo(BigDecimal.ZERO) <= 0) {
					return BigDecimal.ONE.negate();
				} else {
					probabilities_nn[index] = new BigDecimal(factors[index])
							.divide(costs.get(eventName),
									BaseSolverConstants.BIGDECIMAL_SCALE,
									RoundingMode.CEILING);
					reciprocalCostSum = reciprocalCostSum
							.add(probabilities_nn[index]);
					index++;
				}
			}

			BigDecimal eventSum = BigDecimal.valueOf(events.getEventSum());
			BigDecimal quality = BigDecimal.ZERO;
			index = 0;
			for (String eventName : EventConstants.getEvents()) {
				BigDecimal probability = probabilities_nn[index].divide(
						reciprocalCostSum,
						BaseSolverConstants.BIGDECIMAL_SCALE,
						RoundingMode.CEILING);
				BigDecimal frequency = new BigDecimal(events.getEvents().get(
						eventName)).divide(eventSum,
						BaseSolverConstants.BIGDECIMAL_SCALE,
						RoundingMode.CEILING);
				quality = quality
						.add(probability.add(frequency.negate()).abs());
				index++;
			}

			return quality;
		} catch (Exception e) {
			return BigDecimal.ONE.negate();
		}
	}

	/**
	 * computes the quality value for the used costs and the given event types
	 * all events
	 * 
	 * @param costs
	 * @param events
	 * @return
	 */
	private static BigDecimal computeQuality1(ReconstructionCosts costs,
			EventType events) {
		try {
			BigDecimal reciprocalCostSum = BigDecimal.ZERO;
			for (BigDecimal cost : costs.values()) {
				if (cost.compareTo(BigDecimal.ZERO) <= 0) {
					return BigDecimal.ONE.negate();
				} else {
					reciprocalCostSum = reciprocalCostSum.add(BigDecimal.ONE
							.divide(cost, BaseSolverConstants.BIGDECIMAL_SCALE,
									RoundingMode.CEILING));
				}
			}
			BigDecimal eventSum = BigDecimal.valueOf(events.getEventSum());
			BigDecimal quality = BigDecimal.ZERO;
			for (String eventName : EventConstants.getEvents()) {
				BigDecimal probability = BigDecimal.ONE.divide(
						costs.get(eventName),
						BaseSolverConstants.BIGDECIMAL_SCALE,
						RoundingMode.CEILING).divide(reciprocalCostSum,
						BaseSolverConstants.BIGDECIMAL_SCALE,
						RoundingMode.CEILING);
				BigDecimal frequency = new BigDecimal(events.getEvents().get(
						eventName)).divide(eventSum,
						BaseSolverConstants.BIGDECIMAL_SCALE,
						RoundingMode.CEILING);
				quality = quality
						.add(probability.add(frequency.negate()).abs());
			}

			return quality;
		} catch (Exception e) {
			return BigDecimal.ONE.negate();
		}
	}

	/**
	 * computes the quality value for the used costs and the given event types
	 * cospeciation and sorting + cospeciation, duplication and hostswitch
	 * 
	 * @param costs
	 * @param events
	 * @return
	 */
	private static BigDecimal computeQuality2(ReconstructionCosts costs,
			EventType events) {
		try {
			BigDecimal reciprocalCostSumHostSpeciation = BigDecimal.ZERO;
			BigDecimal eventSumHostSpeciation = BigDecimal.ZERO;
			BigDecimal reciprocalCostSumParasiteSpeciation = BigDecimal.ZERO;
			BigDecimal eventSumParasiteSpeciation = BigDecimal.ZERO;

			for (BigDecimal cost : costs.values()) {
				if (cost.compareTo(BigDecimal.ZERO) <= 0) {
					return BigDecimal.ONE.negate();
				}
			}

			reciprocalCostSumHostSpeciation = reciprocalCostSumHostSpeciation
					.add(BigDecimal.ONE.divide(costs.getCospeciationCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			eventSumHostSpeciation = eventSumHostSpeciation.add(new BigDecimal(
					events.getCospeciations()));
			reciprocalCostSumHostSpeciation = reciprocalCostSumHostSpeciation
					.add(BigDecimal.ONE.divide(costs.getSortingCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			eventSumHostSpeciation = eventSumHostSpeciation.add(new BigDecimal(
					events.getSortings()));

			reciprocalCostSumParasiteSpeciation = reciprocalCostSumParasiteSpeciation
					.add(BigDecimal.ONE.divide(costs.getCospeciationCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			eventSumParasiteSpeciation = eventSumParasiteSpeciation
					.add(new BigDecimal(events.getCospeciations()));
			reciprocalCostSumParasiteSpeciation = reciprocalCostSumParasiteSpeciation
					.add(BigDecimal.ONE.divide(costs.getDuplicationCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			eventSumParasiteSpeciation = eventSumParasiteSpeciation
					.add(new BigDecimal(events.getDuplications()));
			reciprocalCostSumParasiteSpeciation = reciprocalCostSumParasiteSpeciation
					.add(BigDecimal.ONE.divide(costs.getHostswitchCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			eventSumParasiteSpeciation = eventSumParasiteSpeciation
					.add(new BigDecimal(events.getHostswitches()));

			BigDecimal qualityHostSpeciation = BigDecimal.ZERO;
			BigDecimal qualityParasiteSpeciation = BigDecimal.ZERO;
			for (String eventName : EventConstants.getEvents()) {
				if (eventName.equals(EventConstants.COSPECIATION)
						|| eventName.equals(EventConstants.SORTING)) {
					BigDecimal probability = BigDecimal.ONE.divide(
							costs.get(eventName),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING).divide(
							reciprocalCostSumHostSpeciation,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					BigDecimal frequency = new BigDecimal(events.getEvents()
							.get(eventName)).divide(eventSumHostSpeciation,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					qualityHostSpeciation = qualityHostSpeciation
							.add(probability.add(frequency.negate()).abs());
				}
				if (eventName.equals(EventConstants.COSPECIATION)
						|| eventName.equals(EventConstants.DUPLICATION)
						|| eventName.equals(EventConstants.HOSTSWITCH)) {
					BigDecimal probability = BigDecimal.ONE.divide(
							costs.get(eventName),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING).divide(
							reciprocalCostSumParasiteSpeciation,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					BigDecimal frequency = new BigDecimal(events.getEvents()
							.get(eventName)).divide(eventSumParasiteSpeciation,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					qualityParasiteSpeciation = qualityParasiteSpeciation
							.add(probability.add(frequency.negate()).abs());
				}
			}

			return qualityHostSpeciation.add(qualityParasiteSpeciation);
		} catch (Exception e) {
			return BigDecimal.ONE.negate();
		}
	}

	/**
	 * computes the quality value for the used costs and the given event types
	 * cospeciation and sorting + duplication and hostswitch
	 * 
	 * @param costs
	 * @param events
	 * @return
	 */
	private static BigDecimal computeQuality3(ReconstructionCosts costs,
			EventType events) {
		try {
			BigDecimal reciprocalCostSumHostSpeciation = BigDecimal.ZERO;
			BigDecimal eventSumHostSpeciation = BigDecimal.ZERO;
			BigDecimal reciprocalCostSumParasiteSpeciation = BigDecimal.ZERO;
			BigDecimal eventSumParasiteSpeciation = BigDecimal.ZERO;

			for (BigDecimal cost : costs.values()) {
				if (cost.compareTo(BigDecimal.ZERO) <= 0) {
					return BigDecimal.ONE.negate();
				}
			}

			reciprocalCostSumHostSpeciation = reciprocalCostSumHostSpeciation
					.add(BigDecimal.ONE.divide(costs.getCospeciationCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			eventSumHostSpeciation = eventSumHostSpeciation.add(new BigDecimal(
					events.getCospeciations()));
			reciprocalCostSumHostSpeciation = reciprocalCostSumHostSpeciation
					.add(BigDecimal.ONE.divide(costs.getSortingCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			eventSumHostSpeciation = eventSumHostSpeciation.add(new BigDecimal(
					events.getSortings()));

			reciprocalCostSumParasiteSpeciation = reciprocalCostSumParasiteSpeciation
					.add(BigDecimal.ONE.divide(costs.getDuplicationCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			eventSumParasiteSpeciation = eventSumParasiteSpeciation
					.add(new BigDecimal(events.getDuplications()));
			reciprocalCostSumParasiteSpeciation = reciprocalCostSumParasiteSpeciation
					.add(BigDecimal.ONE.divide(costs.getHostswitchCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			eventSumParasiteSpeciation = eventSumParasiteSpeciation
					.add(new BigDecimal(events.getHostswitches()));

			BigDecimal qualityHostSpeciation = BigDecimal.ZERO;
			BigDecimal qualityParasiteSpeciation = BigDecimal.ZERO;
			for (String eventName : EventConstants.getEvents()) {
				if (eventName.equals(EventConstants.COSPECIATION)
						|| eventName.equals(EventConstants.SORTING)) {
					BigDecimal probability = BigDecimal.ONE.divide(
							costs.get(eventName),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING).divide(
							reciprocalCostSumHostSpeciation,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					BigDecimal frequency = new BigDecimal(events.getEvents()
							.get(eventName)).divide(eventSumHostSpeciation,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					qualityHostSpeciation = qualityHostSpeciation
							.add(probability.add(frequency.negate()).abs());
				}
				if (eventName.equals(EventConstants.DUPLICATION)
						|| eventName.equals(EventConstants.HOSTSWITCH)) {
					BigDecimal probability = BigDecimal.ONE.divide(
							costs.get(eventName),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING).divide(
							reciprocalCostSumParasiteSpeciation,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					BigDecimal frequency = new BigDecimal(events.getEvents()
							.get(eventName)).divide(eventSumParasiteSpeciation,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					qualityParasiteSpeciation = qualityParasiteSpeciation
							.add(probability.add(frequency.negate()).abs());
				}
			}

			return qualityHostSpeciation.add(qualityParasiteSpeciation);
		} catch (Exception e) {
			return BigDecimal.ONE.negate();
		}
	}

	/**
	 * computes the quality value for the used costs and the given event types
	 * cospeciation, duplication and hostswitch
	 * 
	 * @param costs
	 * @param events
	 * @return
	 */
	private static BigDecimal computeQuality4(ReconstructionCosts costs,
			EventType events) {
		try {
			for (BigDecimal cost : costs.values()) {
				if (cost.compareTo(BigDecimal.ZERO) <= 0) {
					return BigDecimal.ONE.negate();
				}
			}

			BigDecimal reciprocalCostSum = BigDecimal.ZERO;
			reciprocalCostSum = reciprocalCostSum
					.add(BigDecimal.ONE.divide(costs.getCospeciationCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			reciprocalCostSum = reciprocalCostSum
					.add(BigDecimal.ONE.divide(costs.getDuplicationCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));
			reciprocalCostSum = reciprocalCostSum
					.add(BigDecimal.ONE.divide(costs.getHostswitchCosts(),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));

			BigDecimal eventSum = BigDecimal.valueOf(events.getCospeciations()
					+ events.getDuplications() + events.getHostswitches());
			BigDecimal quality = BigDecimal.ZERO;
			for (String eventName : EventConstants.getEvents()) {
				if (!eventName.equals(EventConstants.SORTING)) {
					BigDecimal probability = BigDecimal.ONE.divide(
							costs.get(eventName),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING).divide(reciprocalCostSum,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					BigDecimal frequency = new BigDecimal(events.getEvents()
							.get(eventName)).divide(eventSum,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
					quality = quality.add(probability.add(frequency.negate())
							.abs());
				}
			}

			return quality;
		} catch (Exception e) {
			return BigDecimal.ONE.negate();
		}
	}

	/**
	 * computes the quality value for the used costs and the given event types
	 * cospeciation+sorting vs. duplication+hostswitch
	 * 
	 * @param costs
	 * @param events
	 * @return
	 */
	private static BigDecimal computeQuality5(ReconstructionCosts costs,
			EventType events) {
		try {
			for (BigDecimal cost : costs.values()) {
				if (cost.compareTo(BigDecimal.ZERO) <= 0) {
					return BigDecimal.ONE.negate();
				}
			}
			BigDecimal costsCS = costs.getCospeciationCosts().add(
					costs.getSortingCosts());
			BigDecimal costsDH = costs.getDuplicationCosts().add(
					costs.getHostswitchCosts());
			BigDecimal eventsCS = BigDecimal.valueOf(events.getCospeciations()
					+ events.getSortings());
			BigDecimal eventsDH = BigDecimal.valueOf(events.getDuplications()
					+ events.getHostswitches());

			BigDecimal reciprocalCostSum = BigDecimal.ONE.divide(costsCS,
					BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);
			reciprocalCostSum = reciprocalCostSum
					.add(BigDecimal.ONE.divide(
							costs.getDuplicationCosts().add(
									costs.getHostswitchCosts()),
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING));

			BigDecimal eventSum = BigDecimal.valueOf(events.getEventSum());

			BigDecimal quality = BigDecimal.ZERO;

			BigDecimal probability = BigDecimal.ONE.divide(costsCS,
					BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING)
					.divide(reciprocalCostSum,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
			BigDecimal frequency = eventsCS.divide(eventSum,
					BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);

			quality = quality.add(probability.add(frequency.negate()).abs());

			probability = BigDecimal.ONE.divide(costsDH,
					BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING)
					.divide(reciprocalCostSum,
							BaseSolverConstants.BIGDECIMAL_SCALE,
							RoundingMode.CEILING);
			frequency = eventsDH.divide(eventSum,
					BaseSolverConstants.BIGDECIMAL_SCALE, RoundingMode.CEILING);

			quality = quality.add(probability.add(frequency.negate()).abs());

			return quality;
		} catch (Exception e) {
			return BigDecimal.ONE.negate();
		}
	}
}
